"""Describes the RBF + Linear kernel, a sum of an RBF kernel
and a Linear kernel. This is highly effective in cases where
there is a linear trend but with local deviations."""
from abc import ABC
from math import ceil

import numpy as np
from scipy.stats import chi

from xGPR.xgpr_cpu_rfgen_cpp_ext import cpuRBFFeatureGen, cpuRBFGrad
try:
    import cupy as cp
    from xGPR.xgpr_cuda_rfgen_cpp_ext import cudaRBFFeatureGen, cudaRBFGrad
except:
    pass
from ..kernel_baseclass import KernelBaseclass



class RBFLinear(KernelBaseclass, ABC):
    """An implementation of a Linear kernel + an RBF kernel.
    Since it inherits from KernelBaseclass, it includes the
    attributes of that class. Only additional attributes unique
    to this class are described here.

    Attributes:
        nblocks (int): The SORF transform is performed in blocks
            that result from H D1 H D2 H D3, where H is the
            normalized Hadamard matrix and D1, D2, D3 are diagonal
            matrices whose elements are drawn from a Rademacher
            distribution. nblocks is the number of such operations
            required given the input size and number of random
            features requested.
        padded_dims (int): The next largest power of two greater than
            xdim[-1], since the Hadamard transform only operates on
            vectors whose length is a power of two.
        internal_rffs (int): The number of random features that will be generated.
            This is different than num_rffs (from the parent class), which is what
            the kernel will report to anyone asking how many features it generates.
            The reason for this difference is that the linear + rbf kernel
            concatenates the input (for the linear portion of the kernel) to
            the random features generated for the RBF portion.
        num_freqs (int): The number of frequencies to sample. Note that this is
            calculated based on internal_rffs not num_rffs so the calculation
            performed by the parent class is overriden.
        radem_diag: The diagonal matrices for the SORF transform. Type is int8.
        chi_arr: A diagonal array whose elements are drawn from the chi
            distribution. Ensures the marginals of the matrix resulting
            from S H D1 H D2 H D3 are correct.
        feature_gen: A reference to the wrapped C function that generates
            features.
        gradfun: A reference to the wrapped C function that calculates
            gradients.
    """

    def __init__(self, xdim, num_rffs, random_seed = 123, device = "cpu",
                num_threads = 2, double_precision = False,
                kernel_spec_parms = {}):
        """Constructor. Calls the KernelBaseclass
        constructor first.

        Args:
            num_rffs (int): The user-requested number of random Fourier features.
                For sine-cosine kernels (RBF, Matern), this will be saved by the
                class as num_rffs. Note that the number of input features (from xdim)
                is subtracted from this to generate internal_rffs. If the result is
                less than zero, the kernel will generate an exception upon being
                created.
            xdim (tuple): The dimensions of the input. Either (N, D) or (N, M, D)
                where N is the number of datapoints, D is number of features
                and M is number of timepoints or sequence elements (convolution
                kernels only).
            num_threads (int): The number of threads to use for generating random
                features if running on CPU. If running on GPU, this is ignored.
            random_seed (int): The seed to the random number generator.
            double_precision (bool): If True, generate random features in double precision.
                Otherwise, generate as single precision.
            kernel_spec_parms (dict): A dictionary of other optional kernel settings.

        Raises:
            ValueError: If a non 2d input array dimensionality is supplied.
        """
        if len(xdim) != 2:
            raise ValueError("The dimensionality of the input is inappropriate for "
                        "the kernel you have selected.")

        #Although this IS a sine_cosine kernel, we don't want the parent class
        #to enforce that num_rffs be a multiple of two (only needs to be true
        #for internal rffs), so we set sine_cosine_kernel to False.
        super().__init__(num_rffs, xdim, num_threads = num_threads,
                sine_cosine_kernel = False,
                double_precision = double_precision,
                kernel_spec_parms = kernel_spec_parms)

        self.internal_rffs = num_rffs - xdim[1]
        if self.internal_rffs <= 1 or self.internal_rffs % 2 != 0:
            raise ValueError("For the RBFLinear kernel, the number of 'random' "
                    "features requested includes the number of features in "
                    "the input. So, for example, if the input is a length 100 "
                    "vector and training_rffs is 1000, 900 random features will "
                    "be generated and the input features will be concatenated to "
                    "this to yield 1000 'random' features. The number of "
                    "training and fitting rffs requested should therefore be "
                    "at least num_input_features + 2, and after the input length "
                    "is subtracted, the remainder should be an even number. The number of "
                    "variance_rffs requested is not affected.")
        self.hyperparams = np.ones((2))
        self.bounds = np.asarray([[1e-3,1e1], [1e-6, 1e2]])


        self.num_freqs = int(self.internal_rffs / 2)
        self.padded_dims = 2**ceil(np.log2(max(xdim[-1], 2)))

        radem_array = np.asarray([-1,1], dtype=np.int8)
        rng = np.random.default_rng(random_seed)
        if self.padded_dims < self.num_freqs:
            self.nblocks = ceil(self.num_freqs / self.padded_dims)
        else:
            self.nblocks = 1
        self.radem_diag = rng.choice(radem_array, size=(3, 1,
                self.nblocks * self.padded_dims), replace=True)
        self.chi_arr = chi.rvs(df=self.padded_dims, size=self.num_freqs,
                            random_state = random_seed)

        if not self.double_precision:
            self.chi_arr = self.chi_arr.astype(np.float32)

        self.device = device



    def kernel_specific_set_device(self, new_device):
        """Called by parent class when device is changed. Moves
        some of the object parameters to the appropriate device
        and updates self.cosfunc, self.sinfunc, which are
        convenience references to np.cos / np.sin or cp.cos
        / cp.sin."""
        if new_device == "cpu":
            if not isinstance(self.radem_diag, np.ndarray):
                self.radem_diag = cp.asnumpy(self.radem_diag)
                self.chi_arr = cp.asnumpy(self.chi_arr)
        else:
            self.radem_diag = cp.asarray(self.radem_diag)
            self.chi_arr = cp.asarray(self.chi_arr)



    def kernel_specific_transform(self, input_x, sequence_length = None):
        """Combines the two steps involved in random feature generation
        to generate random features.

        Args:
            input_x: Either a cupy or numpy array containing the input.
            sequence_length: Accepted for consistency with baseclass and
                kernels that use this argument but is not used by this
                class of kernels and is therefore ignored.

        Returns:
            xtrans: A cupy or numpy array containing the generated features.
        """
        xcopy = input_x.copy()
        input_x *= self.hyperparams[1]
        if self.device == "cpu":
            output_x = np.zeros((input_x.shape[0], self.num_rffs), np.float64)
            rf_features = np.zeros((input_x.shape[0], self.internal_rffs), np.float64)
            cpuRBFFeatureGen(input_x, rf_features, self.radem_diag, self.chi_arr,
                self.num_threads, self.fit_intercept)
        else:
            output_x = cp.zeros((input_x.shape[0], self.num_rffs), cp.float64)
            rf_features = cp.zeros((input_x.shape[0], self.internal_rffs), cp.float64)
            cudaRBFFeatureGen(input_x, rf_features, self.radem_diag, self.chi_arr,
                self.fit_intercept)

        output_x[:,:self.internal_rffs] = rf_features
        output_x[:,self.internal_rffs:] = xcopy
        return output_x


    def kernel_specific_set_hyperparams(self):
        """Provided for consistency with baseclass. This
        kernel has no kernel-specific properties that must
        be reset after hyperparameters are changed."""
        return


    def kernel_specific_gradient(self, input_x, sequence_length = None):
        """The gradient for kernel-specific hyperparameters is calculated
        using an array (dz_dsigma) specific to each kernel.

        Args:
            input_x: A cupy or numpy array containing the raw input data.
            sequence_length: Accepted for consistency with baseclass and
                kernels that use this argument but is not used by this
                class of kernels and is therefore ignored.

        Returns:
            output_x: A cupy or numpy array containing the random feature
                representation of the input.
            dz_dsigma: A cupy or numpy array containing the derivative of
                output_x with respect to the kernel-specific hyperparameters.
        """
        if self.device == "cpu":
            output_x = np.zeros((input_x.shape[0], self.num_rffs), np.float64)
            output_grad = np.zeros((input_x.shape[0], self.num_rffs, 1), np.float64)
            rf_features = np.zeros((input_x.shape[0], self.internal_rffs), np.float64)
            rf_grad = np.zeros((input_x.shape[0], self.internal_rffs, 1), np.float64)
            cpuRBFGrad(input_x, rf_features, rf_grad, self.radem_diag, self.chi_arr,
                self.hyperparams[1], self.num_threads, self.fit_intercept)
        else:
            output_x = cp.zeros((input_x.shape[0], self.num_rffs), cp.float64)
            output_grad = cp.zeros((input_x.shape[0], self.num_rffs, 1), cp.float64)
            rf_features = cp.zeros((input_x.shape[0], self.internal_rffs), cp.float64)
            rf_grad = cp.zeros((input_x.shape[0], self.internal_rffs, 1), cp.float64)
            cudaRBFGrad(input_x, rf_features, rf_grad, self.radem_diag, self.chi_arr,
                self.hyperparams[1], self.fit_intercept)


        output_x[:,:self.internal_rffs] = rf_features
        output_x[:,self.internal_rffs:] = input_x
        output_grad[:,:self.internal_rffs,0:1] = rf_grad
        return output_x, output_grad



"""Contains the tools necessary to generate the matrices required for
the preconditioner on either CPU or GPU, using several different
methods."""
try:
    import cupy as cp
    import cupyx
    from cupyx import scipy
    from cupyx.scipy import linalg
except:
    pass
import numpy as np
import scipy

from ..kernels.srht_compressor import SRHTCompressor



def single_pass_gauss(dataset, kernel, q_mat, acc_results, verbose):
    """Runs a single pass over the dataset using matvecs.

    Args:
        dataset: A valid dataset object.
        kernel: A valid kernel object that can generate
            random features.
        q_mat (array): A (num_rffs, rank) array against
            which the random features will be multiplied.
        acc_results (array): A (num_rffs, rank) array
            in which Z^T Z @ q_mat will be stored.
        verbose (bool): Whether to print updates.
    """
    for j, (xdata, ldata) in enumerate(dataset.get_chunked_x_data()):
        xdata = kernel.transform_x(xdata, ldata)
        acc_results += xdata.T @ (xdata @ q_mat)
        if j % 10 == 0 and verbose:
            print(f"Chunk {j} complete.")


def single_pass_srht(dataset, kernel, compressor, acc_results, verbose):
    """Runs a single pass over the dataset using SRHT.

    Args:
        dataset: A valid dataset object.
        kernel: A valid kernel object that can generate
            random features.
        compressor: A valid kernel object that can perform SRHT.
        acc_results (array): A (num_rffs, rank) array
            in which Z^T Z @ q_mat will be stored.
        verbose (bool): Whether to print updates.
    """
    for j, (xdata, ldata) in enumerate(dataset.get_chunked_x_data()):
        xdata = kernel.transform_x(xdata, ldata)
        acc_results += compressor.transform_x(xdata).T @ xdata
        if j % 10 == 0 and verbose:
            print(f"Chunk {j} complete.")



def subsampled_srht(dataset, kernel, compressor, acc_results, verbose,
        sample_frac = 0.1, random_seed = 123):
    """Runs a single pass over the dataset using SRHT, but sampling the
    data. The resulting preconditioner will not be useful for fitting
    but the calculated ratio is a good predictor of the number of
    iterations required to fit the full dataset, so it can determine
    what max_rank should be used to build the actual preconditioner
    for fitting.

    Args:
        dataset: A valid dataset object.
        kernel: A valid kernel object that can generate
            random features.
        compressor: A valid kernel object that can perform SRHT.
        acc_results (array): A (num_rffs, rank) array
            in which Z^T Z @ q_mat will be stored.
        verbose (bool): Whether to print updates.
        sample_frac (float): The fraction of datapoints to
            sample.
        random_seed (int): Seed for the random number generator.
    """
    rng = np.random.default_rng(random_seed)
    for j, (xdata, ldata) in enumerate(dataset.get_chunked_x_data()):
        cutoff = max(int(sample_frac * float(xdata.shape[0])), 1)
        idx = rng.permutation(xdata.shape[0])[:cutoff]
        if ldata is not None:
            xdata = kernel.transform_x(xdata[idx,...], ldata[idx])
        else:
            xdata = kernel.transform_x(xdata[idx,...])
        acc_results += compressor.transform_x(xdata).T @ xdata
        if j % 10 == 0 and verbose:
            print(f"Chunk {j} complete.")



def single_pass_srht_zty(dataset, kernel, compressor, acc_results, z_trans_y,
                        verbose):
    """Runs a single pass over the dataset using SRHT.

    Args:
        dataset: A valid dataset object.
        kernel: A valid kernel object that can generate
            random features.
        compressor: A valid kernel object that can perform SRHT.
        acc_results (array): A (num_rffs, rank) array
            in which Z^T Z @ q_mat will be stored.
        z_trans_y (array): A (num_rffs) array in which the
            product Z^T y will be stored.
        verbose (bool): Whether to print updates.
    """
    y_trans_y = 0.0
    for j, (xin, yin, ldata) in enumerate(dataset.get_chunked_data()):
        xdata, ydata = kernel.transform_x_y(xin, yin, ldata)
        z_trans_y += xdata.T @ ydata
        y_trans_y += ydata.T @ ydata
        acc_results += compressor.transform_x(xdata).T @ xdata
        if j % 10 == 0 and verbose:
            print(f"Chunk {j} complete.")

    return y_trans_y



def initialize_srht_multipass(dataset, rank, kernel, random_state, verbose = False,
                n_passes = 1, get_zty = False):
    """Builds the randomized Nystrom approximation to the inverse
    of (z^T z + lambda), where z is the random features generated
    for dataset, using SRHT on the first pass with subsequent passes over
    the dataset. This is more expensive but can result in a much
    better preconditioner under certain circumstances.

    Args:
        dataset: An OnlineDataset or OfflineDataset containing the raw data.
        rank (int): The desired rank of the preconditioner.
        kernel: A valid kernel object that can generate random features.
        random_state (int): A seed for the random number generator.
        verbose (bool): If True, print updates during construction.
        get_zty (bool): If True, return z_trans_y and y_trans_y to caller. This
            is useful for some hyperparameter tuning methods.

    Returns:
        u_mat (np.ndarray): The eigenvectors of the matrix needed to
            form the preconditioner.
        s_mat (np.ndarray): The eigenvalues of the
            matrix needed to form the preconditioner.
        z_trans_y: Either None or an array containing the product Z^T @ y,
            depending on whether get_zty is True or False.
        y_trans_y: Either None or a float containing y^T y, depending on
            whether get_zty is True or False.
    """
    z_trans_y, y_trans_y = None, None
    if kernel.device == "cpu":
        acc_results = np.zeros((rank, kernel.get_num_rffs()))
        svd_calculator, cho_calculator = np.linalg.svd, np.linalg.cholesky
        tri_solver = scipy.linalg.solve_triangular
        qr_calculator = np.linalg.qr
        if get_zty:
            z_trans_y, y_trans_y = np.zeros((kernel.get_num_rffs())), 0.0
    else:
        mempool = cp.get_default_memory_pool()
        acc_results = cp.zeros((rank, kernel.get_num_rffs()))
        svd_calculator, cho_calculator = cp.linalg.svd, cp.linalg.cholesky
        tri_solver = cupyx.scipy.linalg.solve_triangular
        qr_calculator = cp.linalg.qr
        if get_zty:
            z_trans_y, y_trans_y = cp.zeros((kernel.get_num_rffs())), 0.0

    compressor = SRHTCompressor(rank, kernel.get_num_rffs(),
                random_seed = random_state, device=kernel.device)

    if get_zty:
        y_trans_y = single_pass_srht_zty(dataset, kernel, compressor, acc_results,
                        z_trans_y, verbose)
    else:
        single_pass_srht(dataset, kernel, compressor, acc_results,
                verbose)

    del compressor
    acc_results = acc_results.T

    if kernel.device == "cuda":
        mempool.free_all_blocks()

    for _ in range(n_passes - 1):
        q_mat, r_mat = qr_calculator(acc_results)
        acc_results[:] = 0.0
        del r_mat
        if kernel.device == "cuda":
            mempool.free_all_blocks()

        single_pass_gauss(dataset, kernel, q_mat, acc_results, verbose)

    if kernel.device == "cuda":
        mempool.free_all_blocks()

    norm = float( np.sqrt((acc_results**2).sum())  )

    shift = np.spacing(norm)
    acc_results += shift * q_mat
    q_mat = q_mat.T @ acc_results

    q_mat = cho_calculator(q_mat)

    if kernel.device == "cuda":
        mempool.free_all_blocks()

    acc_results = tri_solver(q_mat, acc_results.T,
                            overwrite_b = True, lower=True).T

    u_mat, s_mat, _ = svd_calculator(acc_results, full_matrices=False)
    s_mat = (s_mat**2 - shift).clip(min=0)

    return u_mat, s_mat, z_trans_y, y_trans_y


def initialize_srht(dataset, rank, kernel, random_state, verbose = False,
                get_zty = False):
    """Builds the randomized Nystrom approximation to the inverse
    of (z^T z + lambda), where z is the random features generated
    for dataset, using SRHT.

    Args:
        dataset: An OnlineDataset or OfflineDataset containing the raw data.
        rank (int): The desired rank of the preconditioner.
        kernel: A valid kernel object that can generate random features.
        random_state (int): A seed for the random number generator.
        verbose (bool): If True, print updates during construction.
        get_zty (bool): If True, return z_trans_y and y_trans_y to caller. This
            is useful for some hyperparameter tuning methods.

    Returns:
        u_mat (np.ndarray): The eigenvectors of the matrix needed to
            form the preconditioner.
        s_mat (np.ndarray): The eigenvalues of the
            matrix needed to form the preconditioner.
        z_trans_y: Either None or an array containing the product Z^T @ y,
            depending on whether get_zty is True or False.
        y_trans_y: Either None or a float containing y^T y, depending on
            whether get_zty is True or False.
    """
    z_trans_y, y_trans_y = None, None
    if kernel.device == "cpu":
        acc_results = np.zeros((rank, kernel.get_num_rffs()))
        svd_calculator, array_sqrt = np.linalg.svd, np.sqrt
        if get_zty:
            z_trans_y, y_trans_y = np.zeros((kernel.get_num_rffs())), 0.0
    else:
        mempool = cp.get_default_memory_pool()
        acc_results = cp.zeros((rank, kernel.get_num_rffs()))
        svd_calculator, array_sqrt = cp.linalg.svd, cp.sqrt
        if get_zty:
            z_trans_y, y_trans_y = cp.zeros((kernel.get_num_rffs())), 0.0

    compressor = SRHTCompressor(rank, kernel.get_num_rffs(),
                random_seed = random_state, device=kernel.device)


    if get_zty:
        y_trans_y = single_pass_srht_zty(dataset, kernel, compressor, acc_results,
                        z_trans_y, verbose)
    else:
        single_pass_srht(dataset, kernel, compressor, acc_results, verbose)

    c_mat = compressor.transform_x(acc_results)
    _, c_s1, c_v1 = svd_calculator(c_mat, full_matrices = False)

    del c_mat, compressor
    if kernel.device == "cuda":
        mempool.free_all_blocks()

    mask = c_s1 < 1e-14
    c_s1 = 1 / array_sqrt(c_s1.clip(min=1e-14))
    c_s1[mask] = 0
    acc_results = acc_results.T @ c_v1.T @ (c_s1[:,None] * c_v1)

    del c_v1, c_s1
    if kernel.device == "cuda":
        mempool.free_all_blocks()

    u_mat, s_mat, _ = svd_calculator(acc_results, full_matrices=False)
    s_mat = s_mat**2
    if get_zty:
        return u_mat, s_mat, z_trans_y, y_trans_y
    return u_mat, s_mat, None, None




def srht_ratio_check(dataset, rank, kernel, random_state, verbose = False,
                sample_frac = 0.1):
    """Runs a fast 'preconditioner construction' using a random sample of
    the data. The resulting preconditioner will not be useful for fitting,
    so the eigenvectors normally needed for the preconditioner are not
    saved, but the eigenvalues can be used to get an estimated upper
    bound on the condition number of the Hessian formed from the sample,
    and this in turn can be used to estimate what max_rank is actually
    needed to build the preconditioner.

    Args:
        dataset: An OnlineDataset or OfflineDataset containing the raw data.
        rank (int): The desired rank of the preconditioner.
        kernel: A valid kernel object that can generate random features.
        random_state (int): A seed for the random number generator.
        verbose (bool): If True, print updates during construction.
        sample_frac (float): The fraction of datapoints to
            sample.

    Returns:
        s_mat (np.ndarray): The eigenvalues of the
            matrix needed for the ratio check.
    """
    if kernel.device == "cpu":
        acc_results = np.zeros((rank, kernel.get_num_rffs()))
        svd_calculator, array_sqrt = np.linalg.svd, np.sqrt
    else:
        mempool = cp.get_default_memory_pool()
        acc_results = cp.zeros((rank, kernel.get_num_rffs()))
        svd_calculator, array_sqrt = cp.linalg.svd, cp.sqrt

    compressor = SRHTCompressor(rank, kernel.get_num_rffs(),
                random_seed = random_state, device=kernel.device)

    subsampled_srht(dataset, kernel, compressor, acc_results, verbose,
            sample_frac, random_state)

    c_mat = compressor.transform_x(acc_results)
    _, c_s1, c_v1 = svd_calculator(c_mat, full_matrices = False)

    del c_mat, compressor
    if kernel.device == "cuda":
        mempool.free_all_blocks()

    mask = c_s1 < 1e-14
    c_s1 = 1 / array_sqrt(c_s1.clip(min=1e-14))
    c_s1[mask] = 0
    acc_results = acc_results.T @ c_v1.T @ (c_s1[:,None] * c_v1)

    del c_v1, c_s1
    if kernel.device == "cuda":
        mempool.free_all_blocks()

    _, s_mat, _ = svd_calculator(acc_results, full_matrices=False)
    s_mat = s_mat**2
    return s_mat



"""Tests cg fitting for both regression and discriminants."""
import sys
import os
import unittest

import numpy as np

sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".."))
from utils.build_test_dataset import build_test_dataset
from utils.build_classification_dataset import build_discriminant_traintest_split
from utils.model_constructor import get_models
from utils.evaluate_model import evaluate_model

#Sets of hyperparameters known to work well for our testing dataset
#that we can use as a default. HPARAM is for regression only.
HPARAM = np.array([np.log(np.sqrt(0.0767)),  np.log(0.358)])
DISCRIM_HPARAM = np.array([0., -0.75])


NUM_RFFS = 8500


class CheckCGFit(unittest.TestCase):
    """Tests the conjugate gradients algorithm."""

    def test_preconditioned_cg(self):
        """Test using preconditioned cg, which should easily fit
        in under 10 epochs."""
        online_data, _ = build_test_dataset(conv_kernel = False)
        cpu_mod, gpu_mod = get_models("RBF", online_data, num_rffs = NUM_RFFS)

        cpu_mod.set_hyperparams(HPARAM, online_data)
        preconditioner, _ = cpu_mod.build_preconditioner(online_data,
            max_rank = 256, method = "srht")

        niter, _ = cpu_mod.fit(online_data,  preconditioner = preconditioner,
                max_iter = 500, run_diagnostics = True,
                tol = 1e-6,  mode = "cg")
        print(f"CPU, niter: {niter}")
        self.assertTrue(niter < 10)

        if gpu_mod is not None:
            gpu_mod.set_hyperparams(HPARAM, online_data)
            preconditioner, _ = gpu_mod.build_preconditioner(online_data,
                max_rank = 256, method = "srht")

            niter, _ = gpu_mod.fit(online_data,  preconditioner = preconditioner,
                max_iter = 500, run_diagnostics = True,
                tol = 1e-6,  mode = "cg")
            print(f"Cuda, niter: {niter}")
            self.assertTrue(niter < 10)


    def test_autoselect_cg(self):
        """Test using cg when the software automatically selects the
        max_rank."""
        online_data, _ = build_test_dataset(conv_kernel = False)
        cpu_mod, gpu_mod = get_models("RBF", online_data, num_rffs = NUM_RFFS)

        cpu_mod.set_hyperparams(HPARAM, online_data)
        niter, _ = cpu_mod.fit(online_data,
                max_iter = 500, run_diagnostics = True,
                tol = 1e-6,  mode = "cg")
        print(f"CPU autoselected preconditioning, niter: {niter}")
        self.assertTrue(niter < 10)

        if gpu_mod is not None:
            gpu_mod.set_hyperparams(HPARAM, online_data)
            niter, _ = gpu_mod.fit(online_data,
                max_iter = 500, run_diagnostics = True,
                tol = 1e-6,  mode = "cg")
            print(f"Cuda autoselected preconditioning, niter: {niter}")
            self.assertTrue(niter < 10)


if __name__ == "__main__":
    unittest.main()



"""This is a convolution kernel that first generates random
features on the input, then feeds this random feature profile
into an RBF kernel, analogous to a 3-layer CNN."""
from math import ceil

import numpy as np
from scipy.stats import chi

from xGPR.xgpr_cpu_rfgen_cpp_ext import cpuConv1dMaxpool
from xGPR.xgpr_cpu_rfgen_cpp_ext import cpuRBFFeatureGen, cpuRBFGrad
try:
    import cupy as cp
    from xGPR.xgpr_cuda_rfgen_cpp_ext import cudaConv1dMaxpool
    from xGPR.xgpr_cuda_rfgen_cpp_ext import cudaRBFFeatureGen, cudaRBFGrad
except:
    pass

from ..kernel_baseclass import KernelBaseclass




class Conv1dTwoLayer(KernelBaseclass):
    """This kernel first runs ReLU activation and global maxpool
    RF generation on the input, then feeds the output of that
    initial operation into an RBF kernel. This is equivalent
    to running the FastConv1d feature extractor on the input
    then running the output of that operation into an xGPR
    model with an RBF kernel. Sometimes using the FastConv
    feature extractor may be more convenient, and in other cases
    using this single-call kernel may be more convenient.

    Attributes:
        hyperparams (np.ndarray): This kernel has two
            hyperparameters: lambda_ (noise)
            and sigma (inverse mismatch tolerance).
        bounds (np.ndarray): The builtin bounds for hyperparameter optimization,
            which can be overriden / reset by user.
        conv_width (int): The width of the convolution kernel.
            This hyperparameter can be set based on experimentation
            or domain knowledge.
        init_rffs (int): The number of random features generated by
            the first layer.
        radem_diag1: The diagonal matrices for the SORF transform for the
            first layer.
        chi_arr1: A diagonal array whose elements are drawn from the chi
            distribution. Ensures the marginals of the matrix resulting
            from S H D1 H D2 H D3 are correct.
        radem_diag2: The diagonal matrices for the SORF transform for the
            second layer.
        chi_arr2: The diagonal array for the SORF transform for the
            second layer.
        num_threads (int): Number of threads to use if running on CPU;
            ignored if running on GPU.
    """

    def __init__(self, xdim, num_rffs, random_seed = 123,
                device = "cpu", num_threads = 2,
                double_precision = False,
                kernel_spec_parms = {}):
        """Constructor.

        Args:
            xdim (tuple): The dimensions of the input. Either (N, D) or (N, M, D)
                where N is the number of datapoints, D is number of features
                and M is number of timepoints or sequence elements (convolution
                kernels only).
            num_rffs (int): The user-requested number of random Fourier features.
                For sine-cosine kernels (RBF, Matern), this will be saved by the
                class as num_rffs.
            random_seed (int): The seed to the random number generator.
            device (str): One of 'cpu', 'cuda'. Indicates the starting device.
            num_threads (int): The number of threads to use for random feature generation
                if running on CPU. If running on GPU, this is ignored.
            double_precision (bool): If True, generate random features in double precision.
                Otherwise, generate as single precision.
            conv_width (int): The width of the convolution to perform.
            kernel_spec_parms (dict): A dictionary of additional kernel-specific
                attributes.

        Raises:
            RuntimeError: A RuntimeError is raised if the dimensions of the input are
                inappropriate given the conv_width.
        """
        if "conv_width" not in kernel_spec_parms:
            raise ValueError("conv_width must be included as a kernel-specific "
                    "parameter if using a sequence kernel.")
        if "init_rffs" not in kernel_spec_parms:
            raise ValueError("init_rffs must be included as a kernel-specific "
                    "parameter if using the 2 layer conv1d kernel.")
        if len(xdim) != 3:
            raise RuntimeError("Tried to initialize a Conv1d kernel with a 2d x-"
                    "array! x should be a 3d array for Conv1d.")

        self.init_rffs = kernel_spec_parms["init_rffs"]
        if self.init_rffs % 2 != 0:
            raise RuntimeError("Number of init rffs should be an even number.")

        super().__init__(num_rffs, xdim, num_threads = 2,
                sine_cosine_kernel = True, double_precision = double_precision,
                kernel_spec_parms = kernel_spec_parms)
        self.hyperparams = np.ones((2))
        self.bounds = np.asarray([[1e-3,5], [1e-6, 1e2]])

        # First, set up the information needed for the first layer.
        rng = np.random.default_rng(random_seed)
        self.conv_width = kernel_spec_parms["conv_width"]

        dim2_no_padding = self.conv_width * xdim[2]
        padded_dims = 2**ceil(np.log2(max(dim2_no_padding, 2)))
        init_calc_featsize = ceil(self.init_rffs / padded_dims) * \
                        padded_dims

        radem_array = np.asarray([-1,1], dtype=np.int8)

        self.radem_diag1 = rng.choice(radem_array, size=(3, 1, init_calc_featsize),
                                replace=True)
        self.chi_arr1 = chi.rvs(df=padded_dims, size=self.init_rffs,
                            random_state = random_seed)


        # Next, set up the information needed for the RBF output layer.
        padded_dims = 2**ceil(np.log2(max(self.init_rffs, 2)))

        if padded_dims < self.num_freqs:
            nblocks = ceil(self.num_freqs / padded_dims)
        else:
            nblocks = 1
        self.radem_diag2 = rng.choice(radem_array, size=(3, 1,
                nblocks * padded_dims), replace=True)
        self.chi_arr2 = chi.rvs(df=padded_dims, size=self.num_freqs,
                            random_state = random_seed)
        if not self.double_precision:
            self.chi_arr1 = self.chi_arr1.astype(np.float32)
            self.chi_arr2 = self.chi_arr2.astype(np.float32)

        self.num_threads = num_threads
        self.device = device


    def kernel_specific_set_device(self, new_device):
        """Called by parent class when device is changed. Moves
        some of the object parameters to the appropriate device."""
        if new_device == "cuda":
            self.radem_diag1 = cp.asarray(self.radem_diag1)
            self.chi_arr1 = cp.asarray(self.chi_arr1)
            self.radem_diag2 = cp.asarray(self.radem_diag2)
            self.chi_arr2 = cp.asarray(self.chi_arr2)
        elif not isinstance(self.radem_diag1, np.ndarray):
            self.radem_diag1 = cp.asnumpy(self.radem_diag1)
            self.chi_arr1 = cp.asnumpy(self.chi_arr1)
            self.radem_diag2 = cp.asnumpy(self.radem_diag2)
            self.chi_arr2 = cp.asnumpy(self.chi_arr2)



    def kernel_specific_set_hyperparams(self):
        """Provided for consistency with baseclass. This
        kernel has no kernel-specific properties that must
        be reset after hyperparameters are changed."""
        return


    def kernel_specific_transform(self, input_x, sequence_length):
        """Performs the feature generation."""
        if sequence_length is None:
            raise ValueError("sequence_length is required for convolution kernels.")
        if input_x.shape[2] != self._xdim[2]:
            raise RuntimeError("Unexpected input shape supplied.")

        sequence_length = sequence_length.astype(np.int32, copy=False)

        if self.device == "cpu":
            featurized_x = np.zeros((input_x.shape[0], self.init_rffs), np.float32)
            cpuConv1dMaxpool(input_x, featurized_x, self.radem_diag1, self.chi_arr1,
                    sequence_length, self.conv_width, self.num_threads)
            xtrans = np.zeros((featurized_x.shape[0], self.num_rffs), np.float64)
            featurized_x *= self.hyperparams[1]
            cpuRBFFeatureGen(featurized_x, xtrans, self.radem_diag2, self.chi_arr2,
                self.num_threads, self.fit_intercept)

        else:
            featurized_x = cp.zeros((input_x.shape[0], self.init_rffs), cp.float32)
            cudaConv1dMaxpool(input_x, featurized_x, self.radem_diag1, self.chi_arr1,
                    sequence_length, self.conv_width)

            xtrans = cp.zeros((featurized_x.shape[0], self.num_rffs), cp.float64)
            featurized_x *= self.hyperparams[1]
            cudaRBFFeatureGen(featurized_x, xtrans, self.radem_diag2, self.chi_arr2,
                self.fit_intercept)

        return xtrans



    def kernel_specific_gradient(self, input_x, sequence_length = None):
        """The gradient for kernel-specific hyperparameters.

        Args:
            input_x: A cupy or numpy array containing the raw input data.
            sequence_length: A numpy or cupy array containing the number of
                elements in each sequence -- so that zero padding can be masked.

        Returns:
            output_x: A cupy or numpy array containing the random feature
                representation of the input.
            dz_dsigma: A cupy or numpy array containing the derivative of
                output_x with respect to the kernel-specific hyperparameters.
        """
        if self.device == "cpu":
            featurized_x = np.zeros((input_x.shape[0], self.init_rffs), np.float32)
            cpuConv1dMaxpool(input_x, featurized_x, self.radem_diag1, self.chi_arr1,
                    sequence_length, self.conv_width, self.num_threads)

            output_x = np.zeros((input_x.shape[0], self.num_rffs), np.float64)
            dz_dsigma = np.zeros((input_x.shape[0], self.num_rffs, 1), np.float64)
            cpuRBFGrad(featurized_x, output_x, dz_dsigma, self.radem_diag2, self.chi_arr2,
                self.hyperparams[1], self.num_threads, self.fit_intercept)
        else:
            featurized_x = cp.zeros((input_x.shape[0], self.init_rffs), cp.float32)
            cudaConv1dMaxpool(input_x, featurized_x, self.radem_diag1, self.chi_arr1,
                    sequence_length, self.conv_width)

            output_x = cp.zeros((input_x.shape[0], self.num_rffs), cp.float64)
            dz_dsigma = cp.zeros((input_x.shape[0], self.num_rffs, 1), cp.float64)
            cudaRBFGrad(featurized_x, output_x, dz_dsigma, self.radem_diag2, self.chi_arr2,
                self.hyperparams[1], self.fit_intercept)
        return output_x, dz_dsigma



#Version number. Updated if generating a new release.
#Otherwise, do not change.
__version__ = "0.4.7"

#Key imports.
from .xgp_regression import xGPRegression
from .data_handling.data_handling_baseclass import DatasetBaseclass
from .data_handling.dataset_builder import build_regression_dataset
from .data_handling.dataset_builder import build_classification_dataset

from .kernel_fgen import KernelFGen

from .static_layers import FastConv1d



"""A conjugate gradients implementation, set up to work with
a batched right-hand side. Unlike cupy's / scipy's it offers us
the ability to easily have multiple outputs if needed."""
import numpy as np
try:
    import cupy as cp
except:
    pass






class GPU_ConjugateGrad:
    """Performs conjugate gradients to find b in Ab = y.
    Used both for fitting and for NMLL calculations via
    SLQ.
    """

    def __init__(self):
        pass


    def _matvec(self, dataset, kernel, vec, matvec):
        """Performs a matvec operation (Z^T Z + lambda**2) vec,
        where Z is the random features for the raw data
        from dataset.

        Args:
            dataset: An OnlineDataset or OfflineDataset object
                with the raw data.
            kernel: A valid kernel object.
            vec (array): The target array.
            matvec (array): The product (Z^T Z + lambda**2) vec.
                This array is modified in-place.
        """
        matvec[:] = 0
        for x, lengths in dataset.get_chunked_x_data():
            Z = kernel.transform_x(x, lengths)
            matvec += Z.T @ (Z @ vec)
        matvec += kernel.get_lambda()**2 * vec



    def fit(self, dataset, kernel, preconditioner,
            resid, maxiter = 200, tol = 1e-4,
            verbose = True,
            nmll_settings = False):
        """Performs conjugate gradients to evaluate (Z^T Z + lambda)^-1 Z^T y,
        where Z is the random features generated for the dataset and lambda
        is the shared noise hyperparameter.

        Args:
            dataset: An OnlineDataset or OfflineDataset with the raw data.
            kernel: A valid kernel object.
            preconditioner: Either None or a valid preconditioner object.
            resid (cp.ndarray): The initial residuals, using starting
                weights of 0 (this CG function always assumes that
                starting weights are all 0).
            maxiter (int): The maximum number of iterations.
            tol (double): The threshold for convergence.
            verbose (bint): If True, print regular updates.
            nmll_settings (bint): If True, save and
                return the alpha and beta values (used for NMLL calcs).
                If False, return the number of iterations and a list of
                losses (for diagnostic purposes).

        Returns:
            xk (cp.ndarray): The result of (Z^T Z + lambda)^-1 Z^T y.
            converged (bint): If False, we did not converge, just ran
                into maxiter.
            niter: The number of iterations.
        """
        converged = False

        target = resid[:,0,:].copy()
        init_norms = cp.linalg.norm(target, axis=0)

        z_k = cp.zeros((resid.shape[0], 2, resid.shape[2]), dtype=cp.float64)
        p_k = cp.zeros((resid.shape[0], 2, resid.shape[2]), dtype=cp.float64)

        alpha = cp.zeros((resid.shape[2]), dtype=cp.float64)
        beta = cp.zeros((resid.shape[2]), dtype=cp.float64)
        alphas, betas, losses = [], [], []

        x_k = cp.zeros((resid.shape[0], resid.shape[2]), dtype=cp.float64)
        w = x_k.copy()

        if preconditioner is None:
            z_k[:,0,:] = resid[:,0,:]
        else:
            z_k[:,0,:] = preconditioner.batch_matvec(resid[:,0,:])
        p_k[:,0,:] = z_k[:,0,:]

        next_col, current_col = 1, 0
        for niter in range(maxiter):
            self._matvec(dataset, kernel, p_k[:,current_col,:], w)
            alpha[:] = (resid[:,current_col,:] *
                    z_k[:,current_col,:]).sum(axis=0) / \
                    (p_k[:,current_col,:] * w).sum(axis=0)
            x_k += alpha[None,:] * p_k[:,current_col,:]
            resid[:,next_col,:] = resid[:,current_col,:] - \
                        alpha[None,:] * w
            err = cp.linalg.norm(resid[:,current_col,:], axis=0) / init_norms

            if preconditioner is None:
                z_k[:,next_col,:] = resid[:,next_col,:]
            else:
                z_k[:,next_col,:] = preconditioner.batch_matvec(resid[:,next_col,:])

            beta[:] = (resid[:,next_col,:] * z_k[:,next_col,:]).sum(axis=0) / \
                    (resid[:,current_col,:] * z_k[:,current_col,:]).sum(axis=0)
            p_k[:,next_col,:] = z_k[:,next_col,:] + beta[None,:] * \
                    p_k[:,current_col,:]

            if nmll_settings:
                alphas.append(alpha.copy())
                betas.append(beta.copy())
            else:
                losses.append(float(err[0]))

            next_col = abs(next_col - 1)
            current_col = abs(current_col - 1)

            if niter % 5 == 0 and verbose:
                print(f"{niter} iterations complete.")
            if err.max() < tol:
                converged = True
                break

        if nmll_settings:
            if len(alphas) > 1:
                alphas, betas = cp.stack(alphas), cp.stack(betas)
            else:
                alphas = alphas[0].reshape(1, alphas[0].shape[0])
                betas = betas[0].reshape(1, betas[0].shape[0])
            alphas, betas = cp.asnumpy(alphas[:,1:]), cp.asnumpy(betas[:,1:])
            return x_k, alphas, betas
        return x_k[:,0], converged, niter + 1, losses





class CPU_ConjugateGrad:
    """Performs conjugate gradients to find b in Ab = y.
    Used both for fitting and for NMLL calculations via
    SLQ.
    """

    def __init__(self):
        pass


    def _matvec(self, dataset, kernel, vec, matvec):
        """Performs a matvec operation (Z^T Z + lambda**2) vec,
        where Z is the random features for the raw data
        from dataset.

        Args:
            dataset: An OnlineDataset or OfflineDataset object
                with the raw data.
            kernel: A valid kernel object.
            vec (array): The target array.
            matvec (array): The product (Z^T Z + lambda**2) vec.
                This array is modified in-place.
        """
        matvec[:] = 0
        for x, lengths in dataset.get_chunked_x_data():
            Z = kernel.transform_x(x, lengths)
            matvec += Z.T @ (Z @ vec)
        matvec += kernel.get_lambda()**2 * vec


    def fit(self, dataset, kernel, preconditioner,
            resid, maxiter = 200, tol = 1e-4,
            verbose = True,
            nmll_settings = False):
        """Performs conjugate gradients to evaluate (Z^T Z + lambda)^-1 Z^T y,
        where Z is the random features generated for the dataset and lambda
        is the shared noise hyperparameter.

        Args:
            dataset: An OnlineDataset or OfflineDataset with the raw data.
            kernel: A valid kernel object.
            preconditioner: Either None or a valid preconditioner object.
            resid (np.ndarray): The initial residuals, using starting
                weights of 0 (this CG function always assumes that
                starting weights are all 0).
            maxiter (int): The maximum number of iterations.
            tol (double): The threshold for convergence.
            verbose (bint): If True, print regular updates.
            nmll_settings (bint): If True, save and
                return the alpha and beta values (used for NMLL calcs).
                If False, return the number of iterations and a list of
                losses (for diagnostic purposes).

        Returns:
            xk (np.ndarray): The result of (Z^T Z + lambda)^-1 Z^T y.
            converged (bint): If False, we did not converge, just ran
                into maxiter.
            niter: The number of iterations.
        """
        converged = False

        target = resid[:,0,:].copy()
        init_norms = np.linalg.norm(target, axis=0)

        z_k = np.zeros((resid.shape[0], 2, resid.shape[2]), dtype=np.float64)
        p_k = np.zeros((resid.shape[0], 2, resid.shape[2]), dtype=np.float64)

        alpha = np.zeros((resid.shape[2]), dtype=np.float64)
        beta = np.zeros((resid.shape[2]), dtype=np.float64)
        alphas, betas, losses = [], [], []

        x_k = np.zeros((resid.shape[0], resid.shape[2]), dtype=np.float64)
        w = x_k.copy()

        if preconditioner is None:
            z_k[:,0,:] = resid[:,0,:]
        else:
            z_k[:,0,:] = preconditioner.batch_matvec(resid[:,0,:])

        p_k[:,0,:] = z_k[:,0,:]

        next_col, current_col = 1, 0
        for niter in range(maxiter):
            self._matvec(dataset, kernel, p_k[:,current_col,:], w)
            alpha[:] = (resid[:,current_col,:] *
                    z_k[:,current_col,:]).sum(axis=0) / \
                    (p_k[:,current_col,:] * w).sum(axis=0)
            x_k += alpha[None,:] * p_k[:,current_col,:]
            resid[:,next_col,:] = resid[:,current_col,:] - \
                        alpha[None,:] * w
            resid[:,next_col,:] = resid[:,current_col,:] - \
                    alpha[None,:] * w
            err = np.linalg.norm(resid[:,current_col,:], axis=0) / init_norms

            if preconditioner is None:
                z_k[:,next_col,:] = resid[:,next_col,:]
            else:
                z_k[:,next_col,:] = preconditioner.batch_matvec(resid[:,next_col,:])
            beta[:] = (resid[:,next_col,:] * z_k[:,next_col,:]).sum(axis=0) / \
                    (resid[:,current_col,:] * z_k[:,current_col,:]).sum(axis=0)
            p_k[:,next_col,:] = z_k[:,next_col,:] + beta[None,:] * \
                    p_k[:,current_col,:]

            if nmll_settings:
                alphas.append(alpha.copy())
                betas.append(beta.copy())
            else:
                losses.append(float(err[0]))

            next_col = abs(next_col - 1)
            current_col = abs(current_col - 1)

            if niter % 5 == 0 and verbose:
                print(f"{niter} iterations complete.")
            if err.max() < tol:
                converged = True
                break

        if nmll_settings:
            if len(alphas) > 1:
                alphas, betas = np.stack(alphas), np.stack(betas)
            else:
                alphas = alphas[0].reshape(1, alphas[0].shape[0])
                betas = betas[0].reshape(1, betas[0].shape[0])
            return x_k, alphas[:,1:], betas[:,1:]
        return x_k[:,0], converged, niter + 1, losses



"""Provides a function for comparing an exact gradient with a
numerical gradient shared and used by all the kernel-specific
tests."""
import sys
import os
import numpy as np
try:
    import cupy as cp
except:
    pass
from scipy.optimize import approx_fprime

sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".."))
from utils.build_test_dataset import build_test_dataset
from utils.model_constructor import get_models


RANDOM_STATE = 123


def run_kernelspecific_test(kernel_choice, conv_kernel = False,
                training_rffs = 512, conv_ard_kernel = False,
                averaging = 'none'):
    """Compares a numerical gradient with an exact gradient using
    generic hyperparameters and generic kernel settings."""
    online_data, _ = build_test_dataset(conv_kernel)
    xdata, ydata = online_data._xdata[:2000,...], online_data._ydata[:2000]
    cpu_mod, gpu_mod = get_models(kernel_choice, online_data,
                        num_rffs = training_rffs, conv_ard_kernel = conv_ard_kernel,
                        averaging = averaging)

    eps = np.sqrt(np.finfo(np.float32).eps)

    params = np.log(np.full(cpu_mod.get_hyperparams().shape, 0.1))

    cpu_cost, cpu_grad = cpu_mod.exact_nmll_gradient(params, online_data)
    singlepoint_cost = cpu_mod.exact_nmll(params, online_data)
    num_grad = approx_fprime(params, cpu_mod.exact_nmll,
                            eps, online_data)


    print(f"Kernel: {kernel_choice}")
    print(f"Analytic gradient, cpu:  {cpu_grad}")
    print(f"Numerical gradient, cpu: {num_grad}")
    print(f"Singlepoint cost, cpu:   {singlepoint_cost}")

    cpu_gradcomp = 100 * np.max(np.abs(cpu_grad - num_grad) / np.abs(cpu_grad))
    cpu_gradcomp = cpu_gradcomp < 0.5 or np.max(np.abs(cpu_grad - num_grad)) < 0.1
    cpu_costcomp = 100 * np.abs(cpu_cost - singlepoint_cost) / np.abs(singlepoint_cost)
    cpu_costcomp = cpu_costcomp < 0.25

    if gpu_mod is not None:
        gpu_cost, gpu_grad = gpu_mod.exact_nmll_gradient(params, online_data)
        xdata, ydata = cp.asarray(xdata), cp.asarray(ydata)

        singlepoint_cost = gpu_mod.exact_nmll(params, online_data)
        print(f"Analytic gradient, gpu:  {gpu_grad}")
        print(f"Singlepoint cost, gpu:   {singlepoint_cost}")

        gpu_gradcomp = 100 * np.max(np.abs(gpu_grad - cpu_grad) / np.abs(cpu_grad))
        gpu_gradcomp = gpu_gradcomp < 0.5 or np.max(np.abs(gpu_grad - num_grad)) < 0.1
        gpu_costcomp = 100 * np.abs(gpu_cost - singlepoint_cost) / np.abs(singlepoint_cost)
        gpu_costcomp = gpu_costcomp < 0.25
    else:
        gpu_gradcomp, gpu_costcomp = True, True
    return cpu_gradcomp, cpu_costcomp, gpu_gradcomp, gpu_costcomp



"""Contains useful functions for testing convolution-based operations.
We test generally by comparing to the simplest and most idiot-proof
(i.e. verbose and slow) way of doing the operation in question
to the output from the wrapped C / Cuda modules."""
from math import ceil

import numpy as np

from xGPR.xgpr_cpu_rfgen_cpp_ext import cpuFastHadamardTransform as cFHT


def get_initial_matrices_fht(ndatapoints, kernel_width, aa_dim, num_aas,
            num_freqs, mode, precision = "double"):
    """Supplies the initial matrices that the C / Cuda modules will use."""
    dim2 = 2**ceil(np.log2(kernel_width * aa_dim))
    num_blocks = num_aas - kernel_width + 1

    radem_array = np.asarray([-1,1], dtype=np.int8)
    radem_size = ceil(num_freqs / dim2) * dim2
    random_seed = 123
    rng = np.random.default_rng(random_seed)
    xdata = rng.uniform(low=-10.0,high=10.0, size=(ndatapoints, num_aas, aa_dim))

    if mode.startswith("maxpool"):
        features = np.zeros((ndatapoints, num_freqs))
        s_mat = rng.uniform(size=num_freqs)
    else:
        features = np.zeros((ndatapoints, 2 * num_freqs))
        s_mat = rng.uniform(size=num_freqs)

    radem = rng.choice(radem_array, size=(3, 1, radem_size), replace=True)


    sequence_lengths = rng.integers(low=kernel_width + 1, high=num_aas + 1,
            size=ndatapoints).astype(np.int32)

    if precision == "float":
        return dim2, num_blocks, xdata.astype(np.float32), sequence_lengths,\
                features, s_mat.astype(np.float32), radem
    return dim2, num_blocks, xdata, sequence_lengths, features, s_mat, radem



def get_reshaped_x(xdata, kernel_width, dim2, radem, num_blocks,
                    repeat_num, precision = "double"):
    """This is the slow / simple / stupid way to generate the results of
    FHT-based convolution, for comparison with the output from the wrapped
    CPU / Cuda modules. The latter will be faster, but by using this
    straightforward approach we can ensure the former is correct."""

    norm_constant = np.log2(dim2) / 2
    norm_constant = 1 / (2**norm_constant)

    reshaped_x = np.zeros((xdata.shape[0], num_blocks, dim2))
    fht_func = cFHT
    if precision == "float":
        reshaped_x = reshaped_x.astype(np.float32)

    window_size = xdata.shape[2] * kernel_width
    start = repeat_num * reshaped_x.shape[2]
    end = start + reshaped_x.shape[2]

    for i in range(xdata.shape[1] - kernel_width + 1):
        window = xdata[:,i:i+kernel_width,:]
        reshaped_x[:,i,:window_size] = window.reshape((window.shape[0],
                        window.shape[1] * window.shape[2]))
    reshaped_x *= radem[0:1,:,start:end] * norm_constant
    fht_func(reshaped_x, 1)
    reshaped_x *= radem[1:2,:,start:end] * norm_constant
    fht_func(reshaped_x, 1)
    reshaped_x *= radem[2:3,:,start:end] * norm_constant
    fht_func(reshaped_x, 1)

    return reshaped_x


def get_features(xdata, kernel_width, dim2,
            radem, s_mat, num_freqs, num_blocks, sigma,
            sequence_lengths, precision = "double",
            normalization = 0):
    """Builds the ground truth features using an inefficient
    but easy to troubleshoot approach, for comparison with
    the features generated by the extensions."""
    num_repeats = ceil(num_freqs / dim2)
    counter = 0
    features = np.zeros((xdata.shape[0], 2 * num_freqs))

    for i in range(num_repeats):
        reshaped_x = get_reshaped_x(xdata, kernel_width, dim2, radem,
                                num_blocks, i, precision)

        end_position = min((i + 1) * dim2, num_freqs)
        end_position -= i * dim2
        counter = i * dim2

        for j in range(end_position):
            temp = reshaped_x[:,:,j] * s_mat[counter] * sigma

            for k in range(features.shape[0]):
                cutpoint = sequence_lengths[k] - kernel_width + 1
                scaling_factor = 1
                if normalization == 1:
                    scaling_factor /= np.sqrt(cutpoint)
                elif normalization == 2:
                    scaling_factor /= cutpoint
                features[k:k+1, 2 * counter] = np.cos(temp[k:k+1,:cutpoint]).sum() * scaling_factor
                features[k:k+1, 2 * counter + 1] = np.sin(temp[k:k+1,:cutpoint]).sum() * scaling_factor
            counter += 1

    features *= np.sqrt(1 / float(num_freqs))
    return features


def get_features_with_gradient(xdata, kernel_width, dim2,
            radem, s_mat, num_freqs, num_blocks, sigma,
            sequence_lengths, precision = "double"):
    """Builds the ground truth features and gradient using an inefficient
    but easy to troubleshoot approach, for comparison with
    the features generated by the extensions."""
    num_repeats = ceil(num_freqs / dim2)
    counter = 0
    features = np.zeros((xdata.shape[0], 2 * num_freqs))
    gradient = np.zeros(features.shape)

    for i in range(num_repeats):
        reshaped_x = get_reshaped_x(xdata, kernel_width, dim2, radem,
                                num_blocks, i, precision)
        end_position = min((i + 1) * dim2, num_freqs)
        end_position -= i * dim2
        counter = i * dim2
        for j in range(end_position):
            reshaped_x[:,:,j] *= s_mat[counter]
            temp_arr = reshaped_x[:,:,j] * sigma
            for k in range(features.shape[0]):
                cutpoint = sequence_lengths[k] - kernel_width + 1
                gradient[k:k+1,2 * counter] = np.sum(-np.sin(temp_arr[k:k+1,:cutpoint]) *
                    reshaped_x[k:k+1,:cutpoint,j], axis = 1)
                features[k:k+1,2 * counter] = np.sum(np.cos(temp_arr[k:k+1,:cutpoint]), axis = 1)

                gradient[k:k+1,2 * counter + 1] = np.sum(np.cos(temp_arr[k:k+1,:cutpoint]) *
                    reshaped_x[k:k+1,:cutpoint,j], axis = 1)
                features[k:k+1,2 * counter + 1] = np.sum(np.sin(temp_arr[k:k+1,:cutpoint]), axis = 1)
            counter += 1

    gradient *= np.sqrt(1 / num_freqs)
    features *= np.sqrt(1 / num_freqs)
    return features, gradient



def get_features_maxpool(xdata, kernel_width, dim2,
            radem, s_mat, num_rffs, num_blocks,
            sequence_lengths, precision = "double"):
    """Builds the ground truth features for maxpooling using an inefficient
    but easy to troubleshoot approach, for comparison with
    the features generated by the extensions."""
    num_repeats = ceil(num_rffs / dim2)
    features = np.zeros((xdata.shape[0], num_rffs))
    
    for i in range(num_repeats):
        reshaped_x = get_reshaped_x(xdata, kernel_width, dim2, radem,
                                num_blocks, i)
        start, end = i * dim2, min((i + 1) * dim2, num_rffs)
        reshaped_x = s_mat[None, None, start:end] * reshaped_x[...,:(end-start)]

        for k in range(features.shape[0]):
            cutpoint = sequence_lengths[k] - kernel_width + 1
            features[k:k+1, start:end] = reshaped_x[k:k+1,:cutpoint,:].max(axis=1).clip(min=0)
    return features[:,:num_rffs]



"""Tests using both exact fitting and preconditioned CG with
minimal bayes tuning to ensure that we can achieve expected
performance for all currently implemented kernels. This is an
'all-in-one' workflow test, if it fails, run fitting tests,
tuning tests, preconditioner tests and fht operations tests
as appopriate to determine which component is failing."""
import unittest
import sys
import os

from current_kernel_list import IMPLEMENTED_KERNELS

sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".."))
from utils.build_test_dataset import build_test_dataset
from utils.model_constructor import get_models
from utils.evaluate_model import evaluate_model




class CheckPipeline(unittest.TestCase):
    """An all in one pipeline test."""



    def test_fit_cpu(self):
        """Test on cpu."""
        print("Now running CPU tests. Some of these (primarily any "
                "involving ARD kernels) may take a minute.")
        for kernel_name, (is_conv, exp_score) in IMPLEMENTED_KERNELS.items():
            cg_score, exact_score = fit_and_evaluate(kernel_name, is_conv,
                conv_width = 3, get_var = True, device="cpu")
            self.assertTrue(cg_score > exp_score)
            self.assertTrue(exact_score > exp_score)

    def test_fit_gpu(self):
        """Test on gpu."""
        print("Now running GPU tests.")
        for kernel_name, (is_conv, exp_score) in IMPLEMENTED_KERNELS.items():
            cg_score, exact_score = fit_and_evaluate(kernel_name, is_conv,
                conv_width = 3, get_var = True, device="gpu")
            self.assertTrue(cg_score > exp_score)
            self.assertTrue(exact_score > exp_score)


def fit_and_evaluate(kernel, conv_kernel, conv_width = 3,
            get_var = True, conv_ard_kernel = False,
            training_rffs = 512, cg_fitting_rffs = 8192,
            exact_fitting_rffs = 2048, device = "gpu"):
    """Test on a specified device using preconditioned CG and exact."""
    _, train_dataset = build_test_dataset(conv_kernel = conv_kernel)
    cpu_mod, gpu_mod = get_models(kernel, train_dataset, conv_width,
                            conv_ard_kernel = conv_ard_kernel,
                            num_rffs = training_rffs)
    if device == "gpu":
        if gpu_mod is None:
            #If GPU not available, return immediately.
            return None, None
        else:
            model = gpu_mod
    else:
        model = cpu_mod

    model.verbose = False

    test_dataset, _ = build_test_dataset(conv_kernel = conv_kernel,
            xsuffix = "testxvalues.npy", ysuffix = "testyvalues.npy")
    if kernel == "MiniARD":
        model.tune_hyperparams(train_dataset, n_restarts = 1, tol=1e-2,
            tuning_method = "L-BFGS-B")
    else:
        model.tune_hyperparams_crude(train_dataset)


    print(f"Hyperparams, cpu, {kernel}: {model.get_hyperparams()}")
    model.num_rffs = cg_fitting_rffs

    model.fit(train_dataset, max_iter = 500, tol = 1e-6,  mode = "cg")
    cg_score = evaluate_model(model, train_dataset, test_dataset,
            get_var)

    print(f"CG score, {device}, {kernel}: {cg_score}")

    model.num_rffs = exact_fitting_rffs

    model.fit(train_dataset, mode = "exact")
    exact_score = evaluate_model(model, train_dataset, test_dataset,
            get_var)
    print(f"Exact score, {device}, {kernel}: {exact_score}")

    return cg_score, exact_score




if __name__ == "__main__":
    unittest.main()



"""Builds xGPRegression models with generic kernel parameters for
use in other tests."""
import sys
import copy

from xGPR import xGPRegression as xGPReg
#from xGPR import xGPDiscriminant

RANDOM_STATE = 123


def get_models(kernel_choice, dataset, conv_width = 3, num_rffs = 512,
        conv_ard_kernel = False, averaging = 'none'):
    """Generates a CPU model and a GPU model with generic
    kernel settings."""
    if not conv_ard_kernel:
        split_pts = [21,42,63]
    else:
        split_pts = [8]

    cpu_mod = xGPReg(num_rffs = num_rffs, kernel_choice = kernel_choice,
            variance_rffs = 12, random_seed = RANDOM_STATE, device = "cpu",
            kernel_settings = {"matern_nu":5/2,
                            "conv_width":conv_width, "polydegree":2,
                            "split_points":split_pts, "order":2,
                            "intercept":True, "averaging":averaging,
                            "init_rffs":514})
    if "cupy" not in sys.modules:
        print("Cupy not installed -- skipping the CUDA test.")
        gpu_mod = None
    else:
        gpu_mod = copy.deepcopy(cpu_mod)
        gpu_mod.device = "cuda"
        gpu_mod.set_hyperparams(dataset = dataset)

    cpu_mod.set_hyperparams(dataset = dataset)
    return cpu_mod, gpu_mod



"""Describes the xGPRegression class.

The xGPRegression class provides the tools needed to fit a regression
model and make predictions for new datapoints. It inherits from
ModelBaseclass.
"""
import warnings
import numpy as np
from scipy.optimize import minimize
from .cg_toolkit.cg_tools import CPU_ConjugateGrad

try:
    import cupy as cp
    from .preconditioners.cuda_rand_nys_preconditioners import Cuda_RandNysPreconditioner
    from .cg_toolkit.cg_tools import GPU_ConjugateGrad
except:
    print("CuPy not detected. xGPR will run in CPU-only mode.")

from .constants import constants
from .model_baseclass import ModelBaseclass
from .preconditioners.tuning_preconditioners import RandNysTuningPreconditioner
from .preconditioners.inter_device_preconditioners import InterDevicePreconditioner
from .preconditioners.rand_nys_preconditioners import CPU_RandNysPreconditioner

from .fitting_toolkit.cg_fitting_toolkit import cg_fit_lib_internal
from .fitting_toolkit.exact_fitting_toolkit import calc_weights_exact, calc_variance_exact

from .scoring_toolkit.approximate_nmll_calcs import estimate_logdet
from .scoring_toolkit.nmll_gradient_tools import exact_nmll_reg_grad, calc_gradient_terms
from .scoring_toolkit.probe_generators import generate_normal_probes_gpu
from .scoring_toolkit.probe_generators import generate_normal_probes_cpu
from .scoring_toolkit.exact_nmll_calcs import calc_design_mat, direct_weight_calc
from .scoring_toolkit.bayes_grid import bayes_grid_tuning
from .scoring_toolkit.lb_optimizer import shared_hparam_search
from .scoring_toolkit.alpha_beta_optimizer import optimize_alpha_beta




class xGPRegression(ModelBaseclass):
    """An approximate Gaussian process for regression.
    """

    def __init__(self, num_rffs:int = 256,
                    variance_rffs:int = 16,
                    kernel_choice:str = "RBF",
                    device:str = "cpu",
                    kernel_settings:dict = constants.DEFAULT_KERNEL_SPEC_PARMS,
                    verbose:bool = True,
                    num_threads:int = 2,
                    random_seed:int = 123) -> None:
        """The constructor for xGPRegression. Passes arguments onto
        the parent class constructor.

        Args:
            num_rffs (int): The number of random Fourier features
                to use.
            variance_rffs (int): The number of random Fourier features
                to use for posterior predictive variance (i.e. calculating
                uncertainty on predictions). Defaults to 64.
            kernel_choice (str): The kernel that the model will use.
            device (str): Determines whether calculations are performed on
                'cpu' or 'gpu'. The initial entry can be changed later
                (i.e. model can be transferred to a different device).
                Defaults to 'cpu'.
            kernel_settings (dict): Contains kernel-specific parameters --
                e.g. 'matern_nu' for the nu for the Matern kernel, or 'conv_width'
                for the conv1d kernel.
            verbose (bool): If True, regular updates are printed
                during fitting and tuning. Defaults to True.
            num_threads (int): The number of threads to use for random feature generation
                if running on CPU. If running on GPU, this argument is ignored.
            random_seed (int): The seed to the random number generator.
        """
        super().__init__(num_rffs, variance_rffs,
                        kernel_choice, device = device,
                        kernel_settings = kernel_settings,
                        verbose = verbose, num_threads = num_threads,
                        random_seed = random_seed)



    def predict(self, input_x, sequence_lengths = None, get_var:bool = False,
            chunk_size:int = 2000):
        """Generate a predicted value for each
        input datapoint -- and if desired the variance.

        Args:
            input_x (np.ndarray): The input data. Should be a 2d numpy
                array (if non-convolution kernel) or 3d (if convolution
                kernel).
            sequence_lengths: None if you are using a fixed-vector kernel (e.g.
                RBF) and a 1d array of the number of elements in each sequence /
                nodes in each graph if you are using a graph or Conv1d kernel.
            get_var (bool): If True, return (predictions, variance).
                If False, return predictions.
            chunk_size (int): The number of datapoints to process at
                a time. Lower values limit memory consumption. Defaults
                to 2000.

        Returns:
            If get_var is True, returns (predictions, variance). If False,
            returns predictions. Both are numpy arrays of length N
            for N datapoints.

        Raises:
            ValueError: If the dimesionality or type of the input does
                not match what is expected, or if the model has
                not yet been fitted, a ValueError is raised.
        """
        self.pre_prediction_checks(input_x, sequence_lengths, get_var)
        preds, var = [], []

        lambda_ = self.kernel.get_lambda()

        for i in range(0, input_x.shape[0], chunk_size):
            cutoff = min(i + chunk_size, input_x.shape[0])
            if sequence_lengths is not None:
                xfeatures = self.kernel.transform_x(input_x[i:cutoff,...],
                        sequence_lengths[i:cutoff])
            else:
                xfeatures = self.kernel.transform_x(input_x[i:cutoff,...])

            preds.append((xfeatures * self.weights[None, :]).sum(axis = 1))

            if get_var:
                if self.exact_var_calculation:
                    xfeatures = xfeatures[:,:self.variance_rffs]
                    pred_var = (self.var @ xfeatures.T).T
                else:
                    pred_var = self.var.batch_matvec(xfeatures.T).T
                pred_var = lambda_**2 + lambda_**2 * (xfeatures * pred_var).sum(axis=1)
                var.append(pred_var)

        if self.device == "cuda":
            preds = cp.asnumpy(cp.concatenate(preds))
        else:
            preds = np.concatenate(preds)
        if not get_var:
            return preds * self.trainy_std + self.trainy_mean
        if self.device == "cuda":
            var = cp.asnumpy(cp.concatenate(var))
            mempool = cp.get_default_memory_pool()
            mempool.free_all_blocks()
        else:
            var = np.concatenate(var)
        #There may (rarely) be numerical issues which result in variances
        #less than zero; if so, set those variances to zero.
        var[var < 0] = 0
        return preds * self.trainy_std + self.trainy_mean, var * self.trainy_std**2


    def build_preconditioner(self, dataset, max_rank:int = 512, method:str = "srht"):
        """Builds a preconditioner. The resulting preconditioner object
        can be supplied to fit and used for CG.

        Args:
            dataset: A Dataset object.
            max_rank (int): The maximum rank for the preconditioner, which
                uses a low-rank approximation to the matrix inverse. Larger
                numbers mean a more accurate approximation and thus reduce
                the number of iterations, but make the preconditioner more
                expensive to construct.
            method (str): one of "srht", "srht_2". "srht_2" runs two passes
                over the dataset. For the same max_rank, the preconditioner
                built by "srht_2" will reduce the number of CG iterations by
                25-30% compared with "srht", but it does incur the expense
                of a second pass over the dataset.

        Returns:
            preconditioner: A preconditioner object.
            achieved_ratio (float): The min eigval of the preconditioner over
                lambda, the noise hyperparameter shared between all kernels.
                This value has decent predictive value for assessing how
                well the preconditioner is likely to perform.
        """
        self._run_pre_fitting_prep(dataset, max_rank)
        if self.device == "cuda":
            preconditioner = Cuda_RandNysPreconditioner(self.kernel, dataset, max_rank,
                        self.verbose, self.random_seed, method)
            mempool = cp.get_default_memory_pool()
            mempool.free_all_blocks()
        else:
            preconditioner = CPU_RandNysPreconditioner(self.kernel, dataset, max_rank,
                        self.verbose, self.random_seed, method)
        return preconditioner, preconditioner.achieved_ratio




    def exact_nmll(self, hyperparams, dataset):
        """Calculates the exact negative marginal log likelihood (the model
        'score') using matrix decompositions. Fast for small numbers of random
        features but poor scaling to larger numbers. Will return a default
        large number if numerical instability is encountered (only occurs
        for extreme hyperparameter values).

        Args:
            hyperparams (np.ndarray): A numpy array containing the new
                set of hyperparameters that should be assigned to the kernel.
            dataset: An OnlineDataset or OfflineDataset containing the raw
                data we will use for this evaluation.

        Returns:
            negloglik (float): The negative marginal log likelihood for the
                input hyperparameters.
        """
        self._run_singlepoint_nmll_prep(dataset, exact_method = True)

        self.kernel.set_hyperparams(hyperparams, logspace=True)
        ndatapoints = dataset.get_ndatapoints()
        z_trans_z, z_trans_y, y_trans_y = calc_design_mat(dataset, self.kernel)
        if self.device == "cpu":
            diagfunc, logfunc = np.diag, np.log
        else:
            diagfunc, logfunc = cp.diag, cp.log

        #Direct weight calculation may fail IF the hyperparameters supplied
        #lead to a singular design matrix. This is rare, but be prepared to
        #handle if this problem is encountered.
        try:
            chol_z_trans_z, weights = direct_weight_calc(z_trans_z, z_trans_y,
                    self.kernel)
        except:
            warnings.warn("Near-singular matrix encountered when calculating score for "
                    f"hyperparameters {hyperparams}.")
            return constants.DEFAULT_SCORE_IF_PROBLEM

        nll1 = float(0.5 * (y_trans_y - z_trans_y.T @ weights))
        nll2 = float(logfunc(diagfunc(chol_z_trans_z)).sum())
        negloglik, _ = optimize_alpha_beta(self.kernel.get_lambda(),
                np.array([nll1, nll2]), ndatapoints, self.kernel.get_num_rffs())

        #Direct weight calculation may fail IF the hyperparameters supplied
        #lead to a singular design matrix. This is rare, but be prepared to
        #handle if this problem is encountered.
        if np.isnan(negloglik):
            warnings.warn("Near-singular matrix encountered when calculating score for "
                    f"hyperparameters {hyperparams}.")
            return constants.DEFAULT_SCORE_IF_PROBLEM

        if self.verbose:
            print("Evaluated NMLL.")
        return negloglik



    def exact_nmll_gradient(self, hyperparams, dataset, subsample:float = 1):
        """Calculates the gradient of the negative marginal log likelihood w/r/t
        the hyperparameters for a specified set of hyperparameters using
        exact methods (matrix decompositions). Fast for small numbers of
        random features but poor scaling to larger numbers. Will return a default
        large number if numerical instability is encountered (only occurs
        for extreme hyperparameter values).

        Args:
            hyperparams (np.ndarray): The set of hyperparameters at which
                to calculate the gradient.
            dataset: An Online or OfflineDataset containing the raw data
                we will use for this evaluation.
            subsample (float): A value in the range [0.1,1] that indicates what
                fraction of the training set to use each time the gradient is
                calculated (the same subset is used every time). In general, 1
                will give better results, but using a subsampled subset can be
                a fast way to find the (approximate) location of a good
                hyperparameter set.

        Returns:
            negloglik (float): The negative marginal log likelihood.
            grad (np.ndarray): The gradient of the NMLL w/r/t the hyperparameters.
        """
        self._run_singlepoint_nmll_prep(dataset, exact_method = True)

        init_hparams = self.kernel.get_hyperparams()
        self.kernel.set_hyperparams(hyperparams, logspace=True)
        hparams = self.kernel.get_hyperparams(logspace=False)

        if self.verbose:
            print("Evaluating gradient...")

        z_trans_z, z_trans_y, y_trans_y, dz_dsigma_ty, inner_deriv, nsamples = \
                        calc_gradient_terms(dataset, self.kernel, self.device, subsample)

        #Try-except here since very occasionally, optimizer samples a really
        #terrible set of hyperparameters that with numerical error has resulted in a
        #non-positive-definite design matrix. TODO: Find a good workaround to
        #avoid this whole problem.
        try:
            negloglik, grad, _ = exact_nmll_reg_grad(z_trans_z, z_trans_y,
                        y_trans_y, hparams, nsamples, dz_dsigma_ty,
                        inner_deriv, self.device)
        except Exception as _:
            return constants.DEFAULT_SCORE_IF_PROBLEM, hyperparams - init_hparams

        #Same problem as above, only occasionally the LAPACK routines return
        #nan instead of raising an error.
        if np.isnan(negloglik):
            return constants.DEFAULT_SCORE_IF_PROBLEM, hyperparams - init_hparams
        return float(negloglik), grad



    def approximate_nmll(self, hyperparams, dataset, manual_settings:dict = None):
        """Calculates the approximate negative marginal log likelihood (the model
        'score') using stochastic Lanczos quadrature with preconditioning.
        Slower than exact for very small numbers of random features, but
        much faster for large (much better scaling). Avoids numerical
        stability issues, but as an approximation may contain some error,
        especially if fitting parameters are chosen poorly.

        Args:
            hyperparams (np.ndarray): A numpy array containing the new
                set of hyperparameters that should be assigned to the kernel.
            dataset: An OnlineDataset or OfflineDataset containing the raw
                data we will use for this evaluation.
            manual_settings: Either None (default) or a dict. If None, the function will
                try to automatically choose settings that will give a good
                approximation. This process usually works well but you may
                occasionally be able to achieve a better approximation and / or
                better speed by choosing settings yourself. If manual_settings is
                a dict, it can contain the following settings:
                
                * ``"max_rank"``: The preconditioner rank for approximate NMLL estimation.
                  A larger value may improve estimation accuracy, but at the
                  expense of speed. 512 - 1024 is fine for noisy data, 2048 is
                  better if data is close to noise free.

                * ``"nsamples"``: The number of probes for approximate NMLL estimation. 25 (default)
                  is usually fine.

                * ``"niter"``: The maximum number of iterations for approximate NMLL. Should
                  only really become an issue if CG is failing to fit in under 500 iter (the
                  default) -- unusual, hyperparameters that give rise to this issue are likely
                  poor in most cases anyway.

                * ``"tol"``: The convergence tolerance for approximate NMLL. 1e-6 (default)
                  is usually fine. A tighter tol (e.g. 1e-7) can improve accuracy slightly
                  at the expense of increased cost.

                * ``"preconditioner_mode"``: One of "srht", "srht_2". Determines the mode of
                  preconditioner construction.

        Returns:
            nmll (float): The negative marginal log likelihood for the
                input hyperparameters.
        """
        self._run_singlepoint_nmll_prep(dataset, exact_method = False)

        self.kernel.set_hyperparams(hyperparams, logspace = True)
        if self.verbose:
            print("Now building preconditioner...")

        #If the user supplied manual settings, they want to control preconditioner
        #construction. Build a preconditioner using the default settings modified
        #with any settings they selected.
        settings = constants.default_nmll_params
        if manual_settings is not None:
            for key in settings:
                if key in manual_settings:
                    settings[key] = manual_settings[key]

            if settings["max_rank"] >= self.num_rffs:
                settings["max_rank"] = self.num_rffs - 1

            preconditioner = RandNysTuningPreconditioner(self.kernel, dataset,
                        settings["max_rank"], False, self.random_seed,
                        settings["preconditioner_mode"])
        else:
            preconditioner = self._autoselect_preconditioner(dataset,
                    min_rank = constants.SMALLEST_NMLL_MAX_RANK,
                    max_rank = constants.LARGEST_NMLL_MAX_RANK,
                    always_use_srht2 = True,
                    tuning = True)

        if self.verbose:
            print("Now fitting...")

        if self.device == "cuda":
            mempool = cp.get_default_memory_pool()
            mempool.free_all_blocks()
            cg_operator = GPU_ConjugateGrad()
            resid = cp.zeros((self.kernel.get_num_rffs(), 2, settings["nsamples"] + 1),
                            dtype = cp.float64)
            probes = generate_normal_probes_gpu(settings["nsamples"],
                    self.kernel.get_num_rffs(), self.random_seed, preconditioner)
        else:
            cg_operator = CPU_ConjugateGrad()
            resid = np.zeros((self.kernel.get_num_rffs(), 2,
                            settings["nsamples"] + 1), dtype = np.float64)
            probes = generate_normal_probes_cpu(settings["nsamples"],
                self.kernel.get_num_rffs(), self.random_seed, preconditioner)

        z_trans_y = preconditioner.get_zty()
        y_trans_y = preconditioner.get_yty()

        resid[:,0,0] = z_trans_y / dataset.get_ndatapoints()
        resid[:,0,1:] = probes

        x_k, alphas, betas = cg_operator.fit(dataset, self.kernel,
                    preconditioner, resid, settings["nmll_iter"],
                    settings["nmll_tol"], verbose = False,
                    nmll_settings = True)

        x_k[:,0] *= dataset.get_ndatapoints()

        logdet = estimate_logdet(alphas, betas, self.kernel.get_num_rffs(),
                        preconditioner, self.device)

        nll1 = float(0.5 * (y_trans_y - z_trans_y.T @ x_k[:,0]))

        negloglik, _ = optimize_alpha_beta(self.kernel.get_lambda(),
                np.array([nll1, 0.5 * logdet]), dataset.get_ndatapoints(),
                self.kernel.get_num_rffs())
        if self.verbose:
            print("NMLL evaluation completed.")

        return negloglik



    def fit(self, dataset, preconditioner = None,
            tol:float = 1e-6, max_iter:int = 500,
            mode:str = "cg", suppress_var:bool = False,
            max_rank:int = 3000,
            min_rank:int = 512,
            autoselect_target_ratio:float = 30.,
            always_use_srht2:bool = False,
            run_diagnostics:bool = False):
        """Fits the model after checking that the input data
        is consistent with the kernel choice and other user selections.

        Args:
            dataset: Object of class OnlineDataset or OfflineDataset.
            preconditioner: Either None (default) or a valid Preconditioner 
                (generated by a call to build_preconditioner).
                If None and mode is 'cg', a preconditioner is
                automatically constructed at a max_rank chosen using several
                passes over the dataset. If mode is 'exact', this argument is
                ignored.
            tol (float): The threshold below which iterative strategies (L-BFGS, CG)
                are deemed to have converged.
            max_iter (int): The maximum number of epochs for iterative strategies.
            mode (str): Must be one of "cg", "exact".
                Determines the approach used. If 'exact', self.kernel.get_num_rffs
                must be <= constants.constants.MAX_CLOSED_FORM_RFFS.
            suppress_var (bool): If True, do not calculate variance. Use this when you
                are just testing performance on a validation set and so do not need to
                calculate variance.
            max_rank (int): The largest size to which the preconditioner can be set.
                Ignored if not autoselecting a preconditioner (i.e. if
                mode is 'cg' and preconditioner = None). The default is
                significantly more than should be needed the vast majority of
                the time.
            min_rank (int): The smallest rank to which the preconditioner can be
                set. Ignored if not autoselecting a preconditioner (e.g. if mode is
                not 'cg' or preconditioner is not None). The default is usually fine.
                Consider setting to a smaller number if you always want to use the
                smallest preconditioner possible.
            autoselect_target_ratio (float): The target ratio if choosing the
                preconditioner size via autoselect. Lower values reduce the number
                of iterations needed to fit but increase preconditioner expense.
                Default is usually fine if 40 - 50 iterations are considered
                acceptable.
            always_use_srht2 (bool): If True, always use srht2 for preconditioner
                construction. This will reduce the number of iterations for
                fitting by 30% on average but will increase the time cost
                of preconditioner construction about 150%.
            run_diagnostics (bool): If True, some performance metrics are returned.

        Returns:
            Does not return anything unless run_diagnostics is True.
            n_iter (int): The number of iterations if applicable.
            losses (list): The loss on each iteration (only for mode='cg'; empty list
                otherwise).

        Raises:
            ValueError: The input dataset is checked for validity before tuning is
                initiated, an error is raised if problems are found."""
        self._run_pre_fitting_prep(dataset)
        self.weights, self.var = None, None
        self.exact_var_calculation = True

        if self.verbose:
            print("starting fitting")

        if mode == "exact":
            if self.kernel.get_num_rffs() > constants.MAX_CLOSED_FORM_RFFS:
                raise ValueError("You specified 'exact' fitting, but the number of rffs is "
                        f"> {constants.MAX_CLOSED_FORM_RFFS}.")
            self.weights, n_iter, losses = calc_weights_exact(dataset, self.kernel)

        elif mode == "cg":
            if preconditioner is None:
                preconditioner = self._autoselect_preconditioner(dataset,
                        min_rank = min_rank, max_rank = max_rank,
                        ratio_target = autoselect_target_ratio,
                        always_use_srht2 = always_use_srht2)
            self.weights, n_iter, losses = cg_fit_lib_internal(self.kernel, dataset, tol,
                    max_iter, preconditioner, self.verbose)

        else:
            raise ValueError("Unrecognized fitting mode supplied. Must provide one of "
                        "'cg', 'exact'.")

        if not suppress_var:
            if self.verbose:
                print("Now performing variance calculations...")
            #We use exact variance calc UNLESS we are dealing with a linear kernel with
            #a very large number of input features. Find a better / more satisfactory
            #way to resolve this...
            if "Linear" in self.kernel_choice:
                self.var = InterDevicePreconditioner(self.kernel, dataset,
                        self.variance_rffs, False, self.random_seed, "srht")
                self.exact_var_calculation = False
            else:
                self.var = calc_variance_exact(self.kernel, dataset, self.kernel_choice,
                                self.variance_rffs)


        if self.verbose:
            print("Fitting complete.")
        if self.device == "cuda":
            mempool = cp.get_default_memory_pool()
            mempool.free_all_blocks()

        if run_diagnostics:
            return n_iter, losses



    #############################
    #The next block of functions are used for tuning hyperparameters. We provide
    #several strategies and of course the user can easily create their own
    #(e.g. using Optuna).
    #############################


    def tune_hyperparams_crude(self, dataset, bounds = None,
                    random_seed:int = 123,
                    max_bayes_iter:int = 30,
                    subsample:float = 1):
        """Tunes the hyperparameters using Bayesian optimization, but with
        a 'trick' that simplifies the problem greatly for 1-3 hyperparameter
        kernels. Hyperparameters are scored using an exact NMLL calculation.
        The NMLL calculation uses a matrix decomposition with cubic
        scaling in the number of random features, so it is slow
        for anything more than 3-4,000 random features, but has
        low risk of overfitting and is easy to use. It is therefore
        intended as a "quick-and-dirty" method. We recommend using
        this with a small number of random features (e.g. 500 - 3000)
        and call self.tune_hyperparams to fine-tune further if needed,
        or fine-tune using an outside library (Optuna) or even simple
        grid search.

        Args:
            dataset: Object of class OnlineDataset or OfflineDataset.
            random_seed (int): A random seed for the random
                number generator. Defaults to 123.
            max_bayes_iter (int): The maximum number of iterations of Bayesian
                optimization.
            bounds (np.ndarray): The bounds for optimization. If None, default
                boundaries for the kernel will be used. Otherwise, must be an
                array of shape (# hyperparams, 2).
            subsample (float): A value in the range [0.01,1] that indicates what
                fraction of the training set to use each time the score is
                calculated (the same subset is used every time). In general, 1
                will give better results, but using a subsampled subset can be
                a fast way to find the (approximate) location of a good
                hyperparameter set.

        Returns:
            hyperparams (np.ndarray): The best hyperparams found during optimization.
            n_feval (int): The number of function evaluations during optimization.
            best_score (float): The best negative marginal log-likelihood achieved.

        Raises:
            ValueError: The input dataset is checked for validity before tuning is
                initiated. If problems are found, a raise exception will provide an
                explanation of the error. This method will also raise an exception
                if you try to use it on a kernel with > 4 hyperparameters (since
                this strategy no longer provides any benefit under those conditions).
        """
        if subsample < 0.01 or subsample > 1:
            raise ValueError("subsample must be in the range [0.01, 1].")

        optim_bounds = self._run_pre_nmll_prep(dataset, bounds)
        num_hparams = self.kernel.get_hyperparams().shape[0]
        if num_hparams == 1:
            best_score, hyperparams = shared_hparam_search(np.array([]), self.kernel,
                    dataset, optim_bounds, subsample = subsample)
            n_feval = 1
        elif 4 > num_hparams > 1:
            hyperparams, _, best_score, n_feval = bayes_grid_tuning(self.kernel,
                                dataset, optim_bounds, random_seed, max_bayes_iter,
                                self.verbose, subsample = subsample)

        else:
            raise ValueError("The crude procedure is only appropriate for "
                    "kernels with 1-3 hyperparameters.")

        self.kernel.set_hyperparams(hyperparams, logspace=True)
        return hyperparams, n_feval, best_score


    def tune_hyperparams(self, dataset, bounds = None,
            max_iter:int = 50, tuning_method:str = "Powell",
            starting_hyperparams = None, tol:float = 1e-2,
            n_restarts:int = 1, nmll_method:str = "exact",
            manual_settings:dict = None):
        """Tunes the hyperparameters WITHOUT using gradient information.
        The methods included here can be extremely efficient for one-
        and two-hyperparameter kernels but are not recommended for kernels
        with many hyperparameters. This method can either use exact
        NMLL (small num_rffs) or approximate (scales to larger num_rffs, but
        approximation is only good so long as settings selected are reasonable).

        Args:
            dataset: Object of class OnlineDataset or OfflineDataset.
            bounds (np.ndarray): The bounds for optimization. If None, default
                boundaries for the kernel will be used. Otherwise, must be an
                array of shape (# hyperparams, 2).
            max_iter (int): The maximum number of iterations.
            tuning_method (str): One of 'Powell', 'Nelder-Mead', 'L-BFGS-B'.
                'Nelder-Mead' and 'Powell' do not perform gradient calculations
                so have much better scaling to large numbers of RFFs but may take
                more iterations to converge. 'Nelder-Mead' takes longer
                to converge than 'Powell' but is more precise.
                'L-BFGS-B' uses exact gradient calculations if 'nmll_method=exact',
                which have poor scaling to > 4000 RFFs but can converge quite
                quickly. With 'nmll_method=approximate', it uses a finite
                difference approximation. For a good guideline on which
                method to prefer, see the User Guide.
            starting_hyperparams: Either None or a numpy array of shape (nhparams)
                where (nhparams) is the number of hyperparameters for the selected
                kernel. If None, a default is used.
            tol (float): The largest shift in NMLL before the algorithm is
                asssumed to have converged.
            n_restarts (int): The number of times to restart the optimizer from
                a new random starting point.
            nmll_method (str): One of 'exact', 'approximate'. 'Exact' is very fast
                if num_rffs is small but has cubic scaling with num_rffs. On GPU,
                it can work well up to 8,000 RFFs or so.
                'approximate' is slow but has better scaling (i.e. the
                time taken increases linearly with the number of RFFs).
            manual_settings: Either None (default) or a dict. If None, the function will
                try to automatically choose settings that will give a good
                approximation. This process usually works well but you may
                occasionally be able to achieve a better approximation and / or
                better speed by choosing settings yourself. If manual_settings is
                a dict, it can contain the following settings:
                
                * ``"max_rank"``: The preconditioner rank for approximate NMLL estimation.
                  A larger value may improve estimation accuracy, but at the
                  expense of speed. 512 - 1024 is fine for noisy data, 2048 is
                  better if data is close to noise free.

                * ``"nsamples"``: The number of probes for approximate NMLL estimation. 25 (default) is usually fine.

                * ``"niter"``: The maximum number of iterations for approximate NMLL. Should
                  only really become an issue if CG is failing to fit in under 500 iter (the
                  default) -- unusual, hyperparameters that give rise to this issue are likely
                  poor in most cases anyway.

                * ``"tol"``: The convergence tolerance for approximate NMLL. 1e-6 (default)
                  is usually fine. A tighter tol (e.g. 1e-7) can improve accuracy slightly
                  at the expense of increased cost.

                * ``"preconditioner_mode"``: One of "srht", "srht_2". Determines the mode of
                  preconditioner construction.

        Returns:
            hyperparams (np.ndarray): The best hyperparams found during optimization.
            n_feval (int): The number of function evaluations during optimization.
            best_score (float): The best negative marginal log-likelihood achieved.

        Raises:
            ValueError: A ValueError is raised if invalid inputs are supplied.
        """
        if tuning_method == "Powell":
            options={"maxfev":max_iter, "xtol":1e-1, "ftol":tol}
        elif tuning_method == "Nelder-Mead":
            options={"maxfev":max_iter, "fatol":tol}
        elif tuning_method == "L-BFGS-B":
            if nmll_method == "approximate":
                raise ValueError("Approximate NMLL is not supported for "
                        "L-BFGS-B at this time.")
            options={"maxiter":max_iter, "ftol":tol}
        else:
            raise ValueError("Invalid tuning method supplied.")

        optim_bounds = self._run_pre_nmll_prep(dataset, bounds)

        if nmll_method == "approximate":
            args = (dataset, manual_settings)
            cost_fun = self.approximate_nmll
        elif nmll_method == "exact":
            args = (dataset,)
            if tuning_method == "L-BFGS-B":
                cost_fun = self.exact_nmll_gradient
            else:
                cost_fun = self.exact_nmll
        else:
            raise ValueError("Invalid nmll method supplied.")


        bounds_tuples = list(map(tuple, optim_bounds))
        rng = np.random.default_rng(self.random_seed)

        if starting_hyperparams is None:
            x0 = self.kernel.get_hyperparams()
            if (x0 - optim_bounds[:,0]).min() < 0 or \
                    (optim_bounds[:,1] - x0).min() < 0:
                x0 = optim_bounds.mean(axis=1)
                warnings.warn("The kernel hyperparameters were outside the "
                    "optimization boundaries. The mean of the optimization "
                    "boundaries will be used as a starting point.", UserWarning)

        elif isinstance(starting_hyperparams, np.ndarray) and \
            starting_hyperparams.shape[0] == self.kernel.get_hyperparams().shape[0]:
            x0 = starting_hyperparams
        else:
            raise ValueError("Invalid starting hyperparams were supplied.")

        best_score, n_feval = np.inf, 0

        for _ in range(n_restarts):
            if tuning_method != "L-BFGS-B":
                res = minimize(cost_fun, x0 = x0,
                        options=options, method=tuning_method,
                        args = args, bounds = bounds_tuples)
            else:
                res = minimize(cost_fun, x0 = x0,
                        options=options, method=tuning_method,
                        args = args, bounds = bounds_tuples,
                        jac = True)

            n_feval += res.nfev

            if res.fun < best_score:
                n_feval, hyperparams, best_score = res.nfev, res.x, res.fun
            if self.verbose:
                print(f"Best score: {best_score}")
            x0 = [rng.uniform(low = optim_bounds[j,0],
                    high = optim_bounds[j,1]) for j in range(optim_bounds.shape[0])]
            x0 = np.asarray(x0)

        self.kernel.set_hyperparams(hyperparams, logspace=True)
        return hyperparams, n_feval, best_score



"""Describes the bayes_grid optimization routine, which uses a combination
of accelerated gridsearch (to reduce 3d and 2d problems to 1d)
with Bayesian optimization to find good hyperparameters."""
import warnings

import numpy as np
from sklearn.gaussian_process import GaussianProcessRegressor as GPR
from sklearn.gaussian_process.kernels import RBF, Matern

from .lb_optimizer import shared_hparam_search

def bayes_grid_tuning(kernel, dataset, bounds, random_seed,
                    max_iter, verbose, tol = 1e-1,
                    n_pts_per_dim = 100, n_cycles = 1,
                    n_init_pts = 10,
                    subsample = 1):
    """Conducts accelerated gridsearch optimization + Bayesian
    optimization.

    Args:
        kernel: A valid kernel that can generate random features.
        dataset: An OnlineDataset or OfflineDataset containing raw data.
        bounds (np.ndarray): An N x 2 for N hyperparameters set of boundaries
            for optimization.
        random_seed (int): A seed for the random number generator.
        max_iter (int): The maximum number of iterations.
        verbose (bool): If True, print regular updates.
        tol (float): The threshold for convergence.
        n_pts_per_dim (int): The number of grid points per shared hparam.
        n_cycles (int): The number of cycles of "telescoping" grid search
            to run.
        n_init_pts (int): The number of initial grid points to evaluate before
            Bayesian optimization. 10 (the default) is usually fine. If you are
            searcing a smaller space, however, you can sometimes save time by using
            a smaller # (e.g. 5).
        subsample (float): A value in the range [0.01,1] that indicates what
            fraction of the training set to use each time the gradient is
            calculated (the same subset is used every time). In general, 1
            will give better results, but using a subsampled subset can be
            a fast way to find the (approximate) location of a good
            hyperparameter set.

    Raises:
        ValueError: A ValueError is raised if this is run with a kernel with >
            5 or <= 2 hyperparameters.
    """

    if bounds.shape[0] >= 4 or bounds.shape[0] < 2:
        raise ValueError("Bayesian optimization is only allowed for kernels with "
                "2 - 3 hyperparameters.")

    if bounds.shape[0] == 2:
        sigma_grid,_ = get_grid_pts(n_init_pts, bounds)
    else:
        sigma_grid = get_random_starting_pts(n_init_pts, bounds, random_seed)


    sigma_grid = np.round(sigma_grid, 7)

    lb_vals, scores = [], []
    if len(sigma_grid.shape) == 1:
        sigma_grid = sigma_grid.reshape(-1,1)
    sigma_grid = list(sigma_grid)

    for i, sigma_pt in enumerate(sigma_grid):
        score, lb_val = shared_hparam_search(sigma_pt, kernel, dataset, bounds[:1,:],
                        n_pts_per_dim = n_pts_per_dim, n_cycles = n_cycles,
                        subsample = subsample)

        scores.append(score)
        lb_vals.append(lb_val)
        if verbose:
            print(f"Grid point {i} acquired.")

    scores = np.asarray(scores)
    smallest_non_inf_val = np.max(scores[scores < np.inf])
    scores[scores == np.inf] = smallest_non_inf_val
    scores = scores.tolist()
    surrogate = GPR(kernel = RBF(),
            normalize_y = True,
            alpha = 1e-6, random_state = random_seed,
            n_restarts_optimizer = 4)


    sigma_bounds = bounds[1:, :]
    iternum = len(sigma_grid)
    for iternum in range(len(sigma_grid), max_iter):
        new_sigma, min_dist, surrogate = propose_new_point(sigma_grid,
                scores, surrogate, sigma_bounds, random_seed + iternum)
        if verbose:
            print(f"New hparams: {new_sigma}")
        score, lb_val = shared_hparam_search(new_sigma, kernel, dataset, bounds[:1,:],
                        n_pts_per_dim = n_pts_per_dim, n_cycles = n_cycles,
                        subsample = subsample)
        sigma_grid.append(new_sigma)
        lb_vals.append(lb_val)
        scores.append(min(score, smallest_non_inf_val))

        if min_dist < tol:
            break
        if verbose:
            print(f"Additional acquisition {iternum}.")

    best_hparams = np.empty((bounds.shape[0]))
    best_hparams[1:] = sigma_grid[np.argmin(scores)]
    best_hparams[:1]  = lb_vals[np.argmin(scores)]
    if verbose:
        print(f"Best score achieved: {np.round(np.min(scores), 4)}")
        print(f"Best hyperparams: {best_hparams}")
    return best_hparams, (sigma_grid, scores), np.min(scores), iternum


def propose_new_point(sigma_vals, scores,
                surrogate, bounds, random_seed,
                num_cand = 500):
    """Refits the 'surrogate' model and uses it to propose new
    locations for exploration (via Thompson sampling).

    Args:
        sigma_vals (np.ndarray): A grid of kernel-specific hyperparameters
            at which NMLL has been evaluated so far.
        scores (array-like): The scores for sigma_vals.
        surrogate: A scikit-learn Gaussian process model with a Matern kernel.
            Is fitted to sigma_vals, scores.
        bounds (np.ndarray): An N x 2 for N kernel specific hyperparameters
            array of boundaries for optimization.
        random_seed (int): A seed for the random number generator.
        num_cand (int): The number of candidate points to evaluate.

    Returns:
        best_candidate (np.ndarray): A set of kernel-specific hyperparameters
            at which to propose the next acquisition.
        min_dist (float): The smallest distance between best_candidate
            and any existing candidates.
        surrogate: The Gaussian process acquisition model now refitted
            to the updated data.
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        xvals = np.vstack(sigma_vals)
        surrogate.fit(xvals, scores)

    rng = np.random.default_rng(random_seed)
    candidates = rng.uniform(low=bounds[:,0], high=bounds[:,1],
                           size = (num_cand, bounds.shape[0]))
    candidates = np.round(candidates, 7)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        y_candidates = surrogate.sample_y(candidates, n_samples=15,
                    random_state=random_seed)
    best_idx = np.unravel_index(y_candidates.argmin(),
                            y_candidates.shape)
    best_cand = candidates[best_idx[0],:]
    min_dist = np.min(np.linalg.norm(best_cand[None,:] - xvals, axis=1))
    return candidates[best_idx[0],:], min_dist, surrogate


def get_random_starting_pts(num_sigma_vals, bounds, random_seed = 123):
    """For kernels with 4 or 5 hyperparameters, it is often more efficient
    to randomly populate the search space with an initial number of search
    points, and then build on these with Bayesian optimization. This function
    generates the lambda_beta grid plus random 'exploration' values for
    the kernel-specific hyperparameters (aka 'sigma').

    Args:
        num_sigma_vals (int): The number of kernel-specific hyperparameter
            combinations to sample.
        bounds (np.ndarray): The boundaries of the search space.
        random_seed (int): A seed to the random number generator.

    Returns:
        sigma_grid (np.ndarray): The initial kernel-specific hyperparameter combinations to
            evaluate.
        lambda_beta_grid (np.ndarray): The shared hyperparameter combinations to evaluate for each
            sigma combination.
    """
    rng = np.random.default_rng(random_seed)
    sigma_grid = np.empty((num_sigma_vals, bounds.shape[0] - 1))
    for i in range(sigma_grid.shape[1]):
        sigma_grid[:,i] = rng.uniform(size = num_sigma_vals,
                        low = bounds[i + 1, 0], high = bounds[i + 1, 1])
    return sigma_grid


def get_grid_pts(num_pts_per_sigma, bounds):
    """Builds a starting grid of sigma points (kernel-specific
    hyperparameters).
    Args:
        num_pts_per_sigma (int): The number of points at which to
            evaluate NMLL per kernel-specific hyperparameter.
        bounds (np.ndarray): A numpy array of shape N x 2 for N
            hyperparameters indicating the boundaries within which
            to optimize.
    Returns:
        sigma_pts: A numpy array of shape
            (num_pts_per_axis, C) where C is the number of kernel-
            specific hyperparameters. NMLL is evaluated at each
            row in sigma_pts.
    """
    if bounds.shape[0] == 2:
        sigma_pts = np.linspace(bounds[1,0], bounds[1,1],
                                num_pts_per_sigma)
    elif bounds.shape[0] == 3:
        sigma1 = np.linspace(bounds[1,0], bounds[1,1],
                                num_pts_per_sigma)
        sigma2 = np.linspace(bounds[2,0], bounds[2,1],
                                num_pts_per_sigma)
        sigma1, sigma2 = np.meshgrid(sigma1, sigma2)
        sigma_pts = np.array((sigma1.ravel(), sigma2.ravel())).T
    else:
        raise ValueError("This routine is only applicable for "
                    "kernels with < 4 hyperparameters.")
    scoregrid = np.zeros((len(sigma_pts)))
    return sigma_pts, scoregrid



"""Tests the feature generation / gradient calculation routines
specific to MiniARD kernels (MiniARD, MiniARDGraph)."""
import sys
import unittest
import numpy as np
from xGPR.kernels.ARD_kernels.mini_ard import MiniARD

try:
    import cupy as cp
except:
    pass


class TestARDGradCalcs(unittest.TestCase):
    """Runs tests for wrapped ARD gradient calc functions for float and double
    precision for CPU and (if available) GPU. We do this by comparing
    with feature generation using a Python routine that
    makes use of SORFTransform functions, so the basic_fht test must
    pass in order for this test to work."""


    def test_mini_ard_calc(self):
        """Tests gradient calculation for the MiniARD kernel."""
        outcomes = run_mini_ard_grad_test((100,50), 2000, [25])
        for outcome in outcomes:
            self.assertTrue(outcome)

        outcomes = run_mini_ard_grad_test((100,50), 2000, [25],
                fit_intercept = True)
        for outcome in outcomes:
            self.assertTrue(outcome)

        outcomes = run_mini_ard_grad_test((1000,232), 1000, [100,200])
        for outcome in outcomes:
            self.assertTrue(outcome)

        outcomes = run_mini_ard_grad_test((9,2049), 500, [30,450],
                fit_intercept = True)
        for outcome in outcomes:
            self.assertTrue(outcome)





def run_mini_ard_grad_test(xdim, num_freqs, split_points, random_seed = 123,
        fit_intercept = False):
    """A helper function that tests the Cython wrapper which both
    generates RFs and calculates the gradient for specified input params."""
    rng = np.random.default_rng(random_seed)

    input_x = rng.uniform(low=-10.0,high=10.0, size=(xdim[0], xdim[1]))

    gt_double, gt_double_grad = get_MiniARD_gt_values(input_x, num_freqs,
                    split_points, random_seed, double_precision = True,
                    fit_intercept = fit_intercept)

    double_output, double_grad = get_MiniARD_test_values(input_x, num_freqs, split_points,
        random_seed, device = "cpu", double_precision = True,
        fit_intercept = fit_intercept)

    float_output, float_grad = get_MiniARD_test_values(input_x, num_freqs, split_points,
        random_seed, device = "cpu", double_precision = False,
        fit_intercept = fit_intercept)


    if "cupy" in sys.modules:
        cuda_double_output, cuda_double_grad = get_MiniARD_test_values(input_x,
                num_freqs, split_points, random_seed,
                device = "cuda", double_precision = True,
                fit_intercept = fit_intercept)

        cuda_float_output, cuda_float_grad = get_MiniARD_test_values(input_x,
            num_freqs, split_points, random_seed,
            device = "cuda", double_precision = False,
            fit_intercept = fit_intercept)

    outcome_d = np.allclose(gt_double, double_output)
    outcome_f = np.allclose(gt_double, float_output, rtol=1e-5, atol=1e-5)
    outcome_grad_d = np.allclose(gt_double_grad, double_grad)
    outcome_grad_f = np.allclose(gt_double_grad, float_grad, rtol=1e-4, atol=1e-3)
    print("Did the Grad Calc C extension provide the correct result for RBF of "
            f"{xdim}, {num_freqs}, floats, doubles? {outcome_f},{outcome_d}")
    print("Did the Grad Calc C extension provide the correct result for the "
            f"gradient for RBF of {xdim}, {num_freqs}, "
            f"floats, doubles? {outcome_grad_f},{outcome_grad_d}")

    if "cupy" in sys.modules:
        outcome_cuda_d = np.allclose(gt_double, cuda_double_output)
        outcome_cuda_f = np.allclose(gt_double, cuda_float_output, rtol=1e-5, atol=1e-5)
        outcome_cuda_grad_d = np.allclose(gt_double_grad, cuda_double_grad)
        outcome_cuda_grad_f = np.allclose(gt_double_grad, cuda_float_grad, rtol=1e-4, atol=1e-3)
        print("Did the cuda extension provide the correct result for RBF of "
            f"{xdim}, {num_freqs} for floats, doubles? "
            f"{outcome_cuda_f},{outcome_cuda_d}")
        print("Did the Grad Calc cuda extension provide the correct result for the "
            f"gradient for RBF of {xdim}, {num_freqs} for floats, doubles? "
            f"{outcome_cuda_grad_f},{outcome_cuda_grad_d}")
        if np.sum([outcome_d, outcome_f, outcome_cuda_d, outcome_cuda_f,
            outcome_grad_d, outcome_cuda_grad_d]) < 6:
            import pdb
            pdb.set_trace()
        return outcome_d, outcome_f, outcome_cuda_d, outcome_cuda_f, \
                outcome_grad_d, outcome_cuda_grad_d
    return outcome_d, outcome_f, outcome_grad_d, outcome_grad_f



def build_mini_ard_kernel(xdim, num_freqs, split_points, random_seed,
        double_precision = False, fit_intercept = False):
    """In this test, we use the MiniARD kernel to do some of
    the legwork in test setup for us. This function builds
    a MiniARD kernel with some specified values."""
    kernel = MiniARD(xdim, 2 * num_freqs, random_seed,
            device = "cpu", double_precision = double_precision,
            kernel_spec_parms = {"split_points":split_points,
                "intercept":fit_intercept})
    kernel.precompute_weights()
    return kernel


def get_MiniARD_gt_values(input_x, num_freqs, split_points,
        random_seed, double_precision = False, fit_intercept = False):
    """Generates the 'ground-truth' MiniARD grad values for
    comparison with those from the C / Cuda extension. Since
    the MiniARD kernel uses the rbfFeatureGen wrapped functions
    to generate random features when transform_x is called,
    we use these as 'ground truth' for rfs and compare them
    with those generated by the gradient_x function.
    We then calculate the 'ground truth' gradient separately."""
    if double_precision:
        retyped_x = input_x.astype(np.float64)
    else:
        retyped_x = input_x.astype(np.float32)

    kernel = build_mini_ard_kernel(input_x.shape, num_freqs, split_points,
            random_seed, double_precision = double_precision,
            fit_intercept = fit_intercept)
    no_intercept_kernel = build_mini_ard_kernel(input_x.shape, num_freqs, split_points,
            random_seed, double_precision = double_precision,
            fit_intercept = False)
    xtrans = kernel.transform_x(input_x)
    p_weights = kernel.precomputed_weights.copy()
    #This is an ugly hack that compensates for the fact that the scaling factor
    #is slightly different if using an intercept. TODO: Find a cleaner way to
    #do this in the test.
    if fit_intercept:
        xtrans[:,0] = no_intercept_kernel.transform_x(input_x)[:,0]
        xtrans[:,0] /= np.sqrt(2 / (p_weights.shape[0]))
        xtrans[:,0] *= np.sqrt(2 / (p_weights.shape[0] - 0.5))
    gradient = np.zeros((xtrans.shape[0], xtrans.shape[1],
                    len(split_points) + 1))

    ks = kernel.split_pts
    for i in range(ks.shape[0] - 1):
        temp_arr = np.einsum("ij,kj->ki", p_weights[:, ks[i]:ks[i+1]],
                                    retyped_x[:,ks[i]:ks[i+1]])
        for j in range(p_weights.shape[0]):
            gradient[:,2*j,i] = -temp_arr[:,j] * xtrans[:,2*j+1]
            gradient[:,2*j+1,i] = temp_arr[:,j] * xtrans[:,2*j]
    if fit_intercept:
        gradient[:,0,:] = 0
        xtrans[:,0] = kernel.hyperparams[1]
    return xtrans, gradient


def get_MiniARD_test_values(input_x, num_freqs, split_points,
        random_seed, device = "cpu", double_precision = False,
        fit_intercept = False):
    """Generates the test values for comparison with the ground-
    truth values. In this case, we use the MiniARD kernel for
    convenience."""
    kernel = build_mini_ard_kernel(input_x.shape, num_freqs, split_points,
            random_seed, double_precision = double_precision,
            fit_intercept = fit_intercept)
    kernel.device = device
    if device == "cuda":
        x_device = cp.asarray(input_x)
    else:
        x_device = input_x
    xtrans, gradient = kernel.gradient_x(x_device)
    if device == "cuda":
        xtrans = cp.asnumpy(xtrans)
        gradient = cp.asnumpy(gradient)
    return xtrans, gradient



if __name__ == "__main__":
    unittest.main()



"""Implements a randomized nystrom preconditioner for Cuda."""
try:
    import cupy as cp
    from cupyx.scipy.sparse.linalg import LinearOperator as cpx_LinearOperator
except:
    pass
from .rand_nys_constructors import initialize_srht, initialize_srht_multipass



class Cuda_RandNysPreconditioner(cpx_LinearOperator):
    """Implements a preconditioner based on the randomized
    Nystrom approximation to the inverse of Z^T Z + lambda;
    for Cuda only.

    Attributes:
        achieved_ratio (float): lambda_ / min_eig, where min_eig is the
            smallest eigenvalue for the preconditioner. This is a
            reasonably good predictor of how effective the preconditioner
            will be. Smaller is better.
        prefactor (float): A constant by which the matvec is multiplied
            when performing matvecs with the preconditioner.
        u_mat (cp.ndarray): A cupy array containing the eigenvectors of
            the matrix formed during preconditioner construction. Used
            together with prefactor and inv_eig to perform the matvec
            that approximates A^-1.
        inv_eig (cp.ndarray): A cupy array containing the inverse of the
            eigenvalues of the matrix formed during preconditioner construction.
    """

    def __init__(self, kernel, dataset, max_rank, verbose,
                random_state = 123, method = "srht"):
        """Class constructor.

        Args:
            kernel: A valid kernel object that can generate random features.
            dataset: A valid online or offline dataset object that can retrieve
                chunked data.
            max_rank (int): The rank of the preconditioner.
            verbose (bool): If True, print regular updates.
            random_state (int): The random number seed for the random number
                generator.
            method (str): One of "srht", "gauss", "srht_2". Determines the method of
                preconditioner construction.
        """
        super().__init__(shape=(kernel.get_num_rffs(),
                            kernel.get_num_rffs()),
                            dtype=cp.float64)
        if method not in ["srht_2", "srht_3", "srht"]:
            raise ValueError("Unknown method supplied for preconditioner "
                    "construction.")

        if method.startswith("srht_"):
            n_passes = int(method.split("_")[1])
            self.u_mat, s_mat, _, _ = initialize_srht_multipass(dataset, max_rank,
                                kernel, random_state, verbose, n_passes)
        elif method == "srht":
            self.u_mat, s_mat, _, _ = initialize_srht(dataset, max_rank,
                                kernel, random_state, verbose)

        lambda_ = kernel.get_lambda()

        self.eig = s_mat + lambda_**2
        self.inv_eig = self.eig.copy()
        mask = self.eig > 1e-14
        self.inv_eig[mask] = 1 / self.eig[mask]
        self.inv_eig[mask==False] = 0.0
        self.eig[mask==False] = 0.0

        min_eig = s_mat.min()
        self.achieved_ratio = min_eig / lambda_**2
        self.prefactor = float(min_eig + lambda_**2)


    def batch_matvec(self, xvec):
        """Returns a matvec of the preconditioner with a set of input
        vectors xvec."""
        xprod = self.u_mat.T @ xvec
        xprod1 = self.u_mat @ (self.inv_eig[:,None] * self.prefactor * xprod)
        xprod2 = xvec - (self.u_mat @ xprod)
        return xprod2 + xprod1


    def rev_batch_matvec(self, xvec):
        """Returns a matvec of the non-inverted preconditioner with a set of input
        vectors xvec."""
        xprod = self.u_mat.T @ xvec
        xprod1 = self.u_mat @ (self.eig[:,None] * xprod) / self.prefactor
        xprod2 = xvec - (self.u_mat @ xprod)
        return xprod2 + xprod1


    def _matvec(self, xvec):
        """Implements the matvec for a single input vector."""
        xprod = self.u_mat.T @ xvec
        xprod1 = self.u_mat @ (self.inv_eig * self.prefactor * xprod)
        xprod2 = xvec - (self.u_mat @ xprod)
        return xprod2 + xprod1



"""Tests the preconditioners, using a dataset we have used
in the past, to ensure that we can achieve a beta / lambda**2
less than an expected value."""
import unittest
import os
import sys

import numpy as np

sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".."))
from utils.build_test_dataset import build_test_dataset
from utils.model_constructor import get_models


#A set of hyperparameters known to work well for our testing dataset
#that we can use as a default.
HPARAM = np.array([-1.39209982,  -1.00860899])


class CheckPreconditioners(unittest.TestCase):
    """Constructs an SRHT and Gaussian preconditioner and ensures both
    can achieve a beta / lambda_**2 as good as past results."""

    def test_srht_preconditioner(self):
        """Constructs an SRHT preconditioner and ensures it
        can achieve a beta / lambda_**2 similar to expected."""
        print("**********Testing SRHT**************")
        online_data, _ = build_test_dataset(conv_kernel = False)
        cpu_mod, gpu_mod = get_models("RBF", online_data, num_rffs=4100)
        cpu_mod.set_hyperparams(HPARAM, online_data)
        _, cpu_ratio = cpu_mod.build_preconditioner(online_data,
            max_rank = 256, method = "srht")
        print(f"CPU ratio: {cpu_ratio}")
        self.assertTrue(cpu_ratio < 0.3)

        #If CUDA is available...
        if gpu_mod is not None:
            gpu_mod.set_hyperparams(HPARAM, online_data)
            _, gpu_ratio = gpu_mod.build_preconditioner(online_data,
                max_rank = 256, method = "srht")
            print(f"GPU ratio: {gpu_ratio}")
            self.assertTrue(gpu_ratio < 0.3)
    

    def test_srht2_preconditioner(self):
        """Constructs an SRHT2 preconditioner and ensures it
        can achieve a beta / lambda_**2 similar to expected."""
        print("**********Testing SRHT2**************")

        online_data, _ = build_test_dataset(conv_kernel = False)
        cpu_mod, gpu_mod = get_models("RBF", online_data, num_rffs=4100)
        cpu_mod.set_hyperparams(HPARAM, online_data)
        _, cpu_ratio = cpu_mod.build_preconditioner(online_data,
            max_rank = 256, method = "srht_2")
        print(f"CPU ratio: {cpu_ratio}")
        self.assertTrue(cpu_ratio < 0.4)

        #If CUDA is available...
        if gpu_mod is not None:
            gpu_mod.set_hyperparams(HPARAM, online_data)
            _, gpu_ratio = gpu_mod.build_preconditioner(online_data,
                max_rank = 256, method = "srht_2")
            print(f"GPU ratio: {gpu_ratio}")
            self.assertTrue(gpu_ratio < 0.4)



    def test_sampled_preconditioner(self):
        """Constructs a sampled preconditioner and determines whether
        the estimated ratio compares favorably to exact."""
        print("**********Testing sampled preconditioner**************")
        online_data, _ = build_test_dataset(conv_kernel = False)
        cpu_mod, gpu_mod = get_models("RBF", online_data, num_rffs=4100)
        cpu_mod.set_hyperparams(HPARAM, online_data)

        _, cpu_ratio = cpu_mod.build_preconditioner(online_data,
            max_rank = 64, method = "srht")
        sampled_ratio = cpu_mod._check_rank_ratio(online_data, 0.5,
                max_rank = 64)
        self.assertTrue((sampled_ratio / cpu_ratio) < 1.5)
        print(f"CPU exact {cpu_ratio}, sampled {sampled_ratio}")

        #If CUDA is available...
        if gpu_mod is not None:
            gpu_mod.set_hyperparams(HPARAM, online_data)
            _, gpu_ratio = gpu_mod.build_preconditioner(online_data,
                max_rank = 64, method = "srht")
            sampled_ratio = gpu_mod._check_rank_ratio(online_data, 0.5,
                    max_rank = 64)
            print(f"GPU exact {gpu_ratio}, sampled {sampled_ratio}")
            self.assertTrue((sampled_ratio / gpu_ratio) < 1.5)


if __name__ == "__main__":
    unittest.main()



"""Tests cg fitting to ensure success both with and without
preconditioning."""
import sys
import os
import unittest

import numpy as np

sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".."))
from utils.build_test_dataset import build_test_dataset
from utils.model_constructor import get_models
from utils.evaluate_model import evaluate_model

#A set of hyperparameters known to work well for our testing dataset
#that we can use as a default.
HPARAM = np.array([np.log(0.0767),  np.log(0.358)])

NUM_RFFS = 2100
ERROR_MARGIN = 1.0
EASY_HPARAMS = np.array([0.,  1.0])
HARD_HPARAMS = np.array([np.log(1e-3),  1.0])

class CheckApproximateNMLL(unittest.TestCase):
    """Checks whether the approximate NMLL is a reasonably
    good approximation to the exact NMLL."""

    def test_approximate_preconditioned_nmll(self):
        """Test the approximate nmll function using preconditioning."""
        online_data, _ = build_test_dataset(conv_kernel = False)
        cpu_mod, gpu_mod = get_models("RBF", online_data, num_rffs = NUM_RFFS)

        outcome = run_exact_approx_comparison(cpu_mod, EASY_HPARAMS, online_data,
                                "cpu")
        print("Preconditioned test with 'easy' hyperparameters for CPU: "
                f"result within {ERROR_MARGIN} percent? {outcome}")
        self.assertTrue(outcome)

        outcome = run_exact_approx_comparison(cpu_mod, HARD_HPARAMS, online_data,
                                "cpu")
        print("Preconditioned test with 'hard' hyperparameters for CPU: "
                f"result within {ERROR_MARGIN} percent? {outcome}")
        self.assertTrue(outcome)


        if gpu_mod is not None:
            outcome = run_exact_approx_comparison(gpu_mod, EASY_HPARAMS, online_data,
                                "gpu")
            print("Preconditioned test with 'easy' hyperparameters for GPU: "
                f"result within {ERROR_MARGIN} percent? {outcome}")
            self.assertTrue(outcome)

            outcome = run_exact_approx_comparison(gpu_mod, HARD_HPARAMS, online_data,
                                "gpu")
            print("Preconditioned test with 'hard' hyperparameters for GPU: "
                f"result within {ERROR_MARGIN} percent? {outcome}")
            self.assertTrue(outcome)


    def test_approximate_nmll(self):
        """Test the approximate nmll function, NO preconditioning."""
        online_data, _ = build_test_dataset(conv_kernel = False)
        cpu_mod, gpu_mod = get_models("RBF", online_data,
                        num_rffs = NUM_RFFS)



def run_exact_approx_comparison(model, hyperparams, dataset, device):
    """A 'helper' function for generating and comparing exact and
    approximate NMLL."""
    exact_nmll = model.exact_nmll(hyperparams, dataset)
    approx_nmll = model.approximate_nmll(hyperparams, dataset)
    outcome = 100 * abs(approx_nmll - exact_nmll) / exact_nmll < ERROR_MARGIN
    print(f"Exact: {exact_nmll}, Approx: {approx_nmll}")
    return outcome


if __name__ == "__main__":
    unittest.main()



"""The baseclass for all other standard kernels and for SORFKernelBaseclass.

KernelBaseclass houses methods and attributes used by all kernels, including
for resetting the device, calculating the gradient for hyperparameters that
all kernels have etc. It also includes some abstractmethods to ensure
each kernel exposes the same methods to the model classes.
"""
import abc
from abc import ABC

import numpy as np
try:
    import cupy as cp
except:
    pass


class KernelBaseclass(ABC):
    """The baseclass for all other kernel classes.

    Attributes:
        num_freqs (int): The number of sampled frequencies.
        num_rffs (int): The number of random fourier features. Note that
            this will be the same for most kernels. The exception is "sine-
            cosine kernels", e.g. RBF, Matern, which use both the sine
            and cosine of each feature to generate two separate random
            features. In this case the user-requested num_rffs must
            be an even number and num_rffs will be set equal to this
            value, while num_freqs will be 0.5 * num_features.
        _xdim (tuple): The dimensionality of the input. Only elements
            1 and onwards are used -- element 0 is the number of datapoints
            and is not used by the kernel, so any value for element 0
            is acceptable.
        device (str): Must be one of 'cpu', 'cuda'. Determines where
            calculations are performed.
        hyperparams (np.ndarray): An array of shape (N) for N hyperparameters.
            Initialized here to None, each child class will set to its
            own defaults.
        bounds (np.ndarray): An array of shape (N, 2) for N hyperparameters
            that determines the optimization bounds for hyperparameter tuning.
            Each kernel has its own defaults. The parent class initializes to
            None, child classes must set this value.
        double_precision (bool): If True, generate random features in double precision.
            Otherwise, generate as single precision.
        fit_intercept (bool): Whether to fit a y-intercept. Defaults to True but can
            be set to False by adding "intercept":False to kernel_spec_parms.
    """

    def __init__(self, num_rffs, xdim, num_threads = 2,
            sine_cosine_kernel = False, double_precision = False,
            kernel_spec_parms = {}):
        """Constructor for the KernelBaseclass.

        Args:
            num_rffs (int): The user-requested number of random Fourier features.
                For sine-cosine kernels (RBF, Matern), this will be saved by the
                class as num_rffs.
            xdim (tuple): The dimensions of the input. Either (N, D) or (N, M, D)
                where N is the number of datapoints, D is number of features
                and M is number of timepoints or sequence elements (convolution
                kernels only).
            num_threads (int): The number of threads to use if running on CPU. If
                running on GPU, this is ignored.
            sine_cosine_kernel (bool): If True, the kernel is a sine-cosine kernel,
                meaning it will sample self.num_freqs frequencies and use the sine
                and cosine of each to generate twice as many features
                (self.num_rffs). sine-cosine kernels only accept num_rffs
                that are even numbers.
            double_precision (bool): If True, generate random features in double precision.
                Otherwise, generate as single precision.
            kernel_spec_parms (dict): A dictionary of kernel settings / kernel-specific
                parameters.

        Raises:
            ValueError: Raises a ValueError if a sine-cosine kernel is requested
                but num_rffs is not an integer multiple of 2.
        """
        self.double_precision = double_precision
        if num_rffs < 2:
            raise ValueError("num_rffs should always be >= 2.")

        if sine_cosine_kernel:
            if not (num_rffs / 2).is_integer():
                raise ValueError("For sine-cosine kernels (e.g. matern, rbf) "
                        "the number of random fourier features must be an integer "
                        "multiple of two.")
            self.num_freqs = int(num_rffs / 2)
            self.num_rffs = num_rffs
        else:
            self.num_freqs = num_rffs
            self.num_rffs = num_rffs

        self.fit_intercept = True
        if "intercept" in kernel_spec_parms:
            if kernel_spec_parms["intercept"] is False:
                self.fit_intercept = False

        self._xdim = xdim
        self.hyperparams = None
        self.bounds = None

        self.num_threads = num_threads



    @abc.abstractmethod
    def kernel_specific_transform(self, input_x, sequence_length):
        """Kernel classes must implement a method that generates
        random features for a given set of inputs."""

    @abc.abstractmethod
    def kernel_specific_gradient(self, input_x, sequence_length):
        """Kernel classes must implement a method that calculates
        the NMLL gradient for any kernel-specific hyperparameters."""

    @abc.abstractmethod
    def kernel_specific_set_device(self, new_device):
        """Kernel classes must implement a method that performs
        any kernel-specific operations needed to switch device
        for the kernel."""

    @abc.abstractmethod
    def kernel_specific_set_hyperparams(self):
        """Kernel classes must implement a method that performs
        any kernel-specific changes necessary after the hyperparameters
        have been reset."""


    def check_bounds(self, bounds):
        """Checks a set of bounds provided by the caller to ensure they
        are valid for the selected kernel.

        Args:
            bounds (np.ndarray): A numpy array of shape (N, 2) where N
                is the number of hyperparameters, with bounds as
                (low, high).

        Raises:
            ValueError: A ValueError is raised if the bounds array
                passed is not valid for this kernel.
        """
        if bounds.shape != self.bounds.shape:
            raise ValueError("You have tried to supply a set of bounds "
                "for hyperparameter tuning that do not match the shape "
                "of the bounds for the kernel you have chosen. The bounds "
                "should be a numpy array of shape [n_hyperparams, 2] "
                "where the first column is the low bound, second column "
                "is the high bound.")


    def get_bounds(self, logspace=True):
        """Returns the bounds for hyperparameter optimization.
        Note that we do not make this a property because callers
        may need to specify whether they are using logspace (or not).

        Args:
            logspace (bool): If True, return the log of the boundary
                values, otherwise, return the actual values.

        Returns:
            self.bounds (np.ndarray): A numpy array of shape
                (N, 2) where N is the number of hyperparameters,
                containing either the (low, high) bounds
                or the natural lof of those bounds.
        """
        if logspace:
            return np.log(self.bounds)
        return self.bounds


    def set_bounds(self, bounds, logspace=True):
        """Sets the bounds for hyperparameter optimization.
        Note that we do not make this a property because callers
        may need to specify whether they are using logspace (or not).

        Args:
            bounds (np.ndarray): Must be an array of shape (N, 2) where
                N is the number of hyperparameters, with bounds as
                (low, high).
            logspace (bool): If True, the supplied values are the log
                of the actual bounds.

        Raises:
            ValueError: A ValueError is raised if the bounds array
                passed by caller is invalid for the chosen kernel.
        """
        self.check_bounds(bounds)
        if logspace:
            self.bounds = np.exp(bounds)
        else:
            self.bounds = bounds


    def validate_new_datapoints(self, input_x):
        """Checks new input data supplied by caller to
        ensure it is compatible with the dimensionality
        of the data used to fit the model.

        Args:
            input_x (np.ndarray): A numpy array containing
                raw input data.

        Returns:
            valid_data (bool): True if input was acceptable,
                False otherwise.
        """
        valid_data = True
        if len(input_x.shape) != len(self._xdim):
            valid_data = False
        elif len(self._xdim) == 3:
            if input_x.shape[2] != self._xdim[2]:
                valid_data = False
            if input_x.shape[1] < 1:
                valid_data = False
        elif input_x.shape[1] != self._xdim[1]:
            valid_data = False
        return valid_data



    def get_hyperparams(self, logspace = True):
        """Returns the kernel hyperparameters. We have not used the property
        decorator since the logspace option is required.

        Args:
            logspace (bool): If True, the natural log of the hyperparameters is
                returned. Defaults to True.

        Returns:
            hyperparameters (np.ndarray): A numpy array of shape (N) for N
                kernel hyperparameters.
        """
        if logspace:
            return np.log(self.hyperparams)
        return self.hyperparams



    def set_hyperparams(self, hyperparams, logspace = True):
        """Sets the kernel hyperparameters.

        Args:
            hyperparams (np.ndarray): A numpy array of shape N for N hyperparameters.
                This function does not check for validity since it is often
                used by optimization algorithms that are merely modifying
                hyperparameters they retrieved from get_hyperparams. A caller
                that is passing values which may be invalid should call
                self.check_hyperparams first.
            logspace (bool): If True, the natural log of the hyperparameters is
                returned. Defaults to True.

        Returns:
            hyperparameters (np.ndarray): A numpy array of shape (N) for N
                kernel hyperparameters.
        """
        if logspace:
            self.hyperparams = np.exp(hyperparams)
        else:
            self.hyperparams = hyperparams
        self.kernel_specific_set_hyperparams()


    def get_lambda(self):
        """For convenience, we enable caller to retrieve only the
        first hyperparameter, which is needed for a variety of
        operations during fitting and tuning. This hyperparameter
        determines the 'noise level' of the data."""
        return self.hyperparams[0]



    def transform_x(self, input_x, sequence_length = None):
        """Given a numpy array as input (and sequence_length,
        which is none for most kernels but must be specified
        for convolution kernels), generate random features
        as output."""
        # This always generates a copy, which means that we
        # are never working on the input data, only on a copy,
        # and can therefore modify it with impunity.
        if not input_x.flags["C_CONTIGUOUS"]:
            if self.double_precision:
                xin = np.ascontiguousarray(input_x, np.float64)
            else:
                xin = np.ascontiguousarray(input_x, np.float32)
        elif self.double_precision:
            xin = input_x.astype(np.float64, copy=True)
        else:
            xin = input_x.astype(np.float32, copy=True)

        if self.device == "cuda":
            xin = cp.asarray(xin)

        slen = None
        if sequence_length is not None:
            slen = sequence_length.astype(np.int32, copy=False)

        xtrans = self.kernel_specific_transform(xin, slen)
        if self.fit_intercept:
            xtrans[:,0] = 1.

        return xtrans



    def transform_x_y(self, input_x, input_y, sequence_length = None):
        """Given a numpy array as input (and sequence_length,
        which is none for most kernels but must be specified
        for convolution kernels), generate random features
        as output. In addition, convert the input y-values
        to live on the same device that the kernel is currently
        on."""
        y_out = input_y
        if self.device == "cuda":
            y_out = cp.asarray(y_out)

        return self.transform_x(input_x, sequence_length), y_out


    def gradient_x(self, input_x, sequence_length = None):
        """Given a numpy array as input (and sequence_length,
        which is none for most kernels but must be specified
        for convolution kernels), generate random features
        and gradient as output."""
        # This always generates a copy, which means that we
        # are never working on the input data, only on a copy,
        # and can therefore modify it with impunity.
        if not input_x.flags["C_CONTIGUOUS"]:
            if self.double_precision:
                xin = np.ascontiguousarray(input_x, np.float64)
            else:
                xin = np.ascontiguousarray(input_x, np.float32)
        elif self.double_precision:
            xin = input_x.astype(np.float64, copy=True)
        else:
            xin = input_x.astype(np.float32, copy=True)

        if self.device == "cuda":
            xin = cp.asarray(xin)

        slen = None
        if sequence_length is not None:
            slen = sequence_length.astype(np.int32, copy=False)

        xtrans, xgrad = self.kernel_specific_gradient(xin, slen)
        if self.fit_intercept:
            xtrans[:,0] = 1.
            if xgrad.shape[2] > 0:
                xgrad[:,0,:] = 0.
        return xtrans, xgrad



    def gradient_x_y(self, input_x, input_y, sequence_length = None):
        """Given a numpy array as input (and sequence_length,
        which is none for most kernels but must be specified
        for convolution kernels), generate random features
        and gradient as output. In addition, convert the input y-values
        to live on the same device that the kernel is currently
        on."""
        y_out = input_y
        if self.device == "cuda":
            y_out = cp.asarray(y_out)

        xtrans, dz_dsigma = self.gradient_x(input_x, sequence_length)
        return xtrans, dz_dsigma, y_out




    def check_hyperparams(self, hyperparams):
        """Checks a suggested set of hyperparameters passed
        by caller to ensure they are valid. This should be used
        when callers want to set the hyperparameters to user-specified
        values. The optimization algorithms bypass this because
        they are merely modifying hyperparameters retrieved from
        the kernel.

        Args:
            hyperparams (np.ndarray): A numpy array of shape (N), where
                N is the number of hyperparameters.

        Raises:
            ValueError: Raises a ValueError if invalid hyperparameters are
                passed.
        """
        if not isinstance(hyperparams, np.ndarray):
            raise ValueError("The starting hyperparameters must be a numpy array.")
        if hyperparams.shape != self.hyperparams.shape:
            raise ValueError(f"The kernel you selected uses {self.hyperparams.shape[0]} "
                "hyperparameters. A hyperparameter array of the incorrect shape was passed.")



    def get_num_rffs(self):
        """Returns number of RFFs. Not a @property because
        we do not want to define a setter, external classes
        should not be able to set."""
        return self.num_rffs

    def get_xdim(self):
        """Returns the _xdim value. No setter since this should
        not be set externally."""
        return self._xdim


    @property
    def device(self):
        """Getter for the device property, which determines
        whether calculations are on CPU or GPU."""
        return self.device_


    @device.setter
    def device(self, value):
        """Setter for device, which determines whether calculations
        are on CPU or GPU. Note that each kernel must also have
        a kernel_specific_set_device function (enforced via
        an abstractmethod) to make any kernel-specific changes
        that occur when the device is switched.

        Args:
            value (str): Must be one of 'cpu', 'cuda'.

        Raises:
            ValueError: A ValueError is raised if an unrecognized
                device is passed.
        """
        if value not in ('cpu', 'cuda'):
            raise ValueError("Unrecognized device supplied. Must be one "
                    "of 'cpu', 'cuda'.")
        self.device_ = value
        self.kernel_specific_set_device(value)



"""Handles optimization for hyperparameters shared across
all kernels using a "telescoping grid" strategy used by
the minimal_bayes and grid optimization schemes,
scoring hyperparameters using NMLL."""
try:
    import cupy as cp
except:
    pass
import numpy as np



def shared_hparam_search(sigma_vals, kernel, dataset, init_bounds,
        n_pts_per_dim = 100, n_cycles = 1, subsample = 1):
    """Uses a "telescoping grid" procedure to search the space of
    shared hyperparameters for a given set of kernel specific hyperparameters
    (sigma_vals), using exact NMLL to score.

    Args:
        sigma_vals (np.ndarray): A set of kernel-specific hyperparameters,
            of shape (# hyperparameters - 2). Should not be shape > 2.
        kernel: A valid kernel object that can generate random features.
        dataset: A valid dataset object that can retrieve chunked data.
        init_bounds (np.ndarray): A 2 x 2 array where [0,:] is the boundaries
            for shared hyperparameter lambda and [1,:] is the boundaries
            for shared hyperparameter beta.
        n_pts_per_dim (int): The number of points per grid dimension.
        n_cycles (int): The number of 'telescoping grid' cycles.
        subsample (float): A value in the range [0.01,1] that indicates what
            fraction of the training set to use each time the gradient is
            calculated (the same subset is used every time). In general, 1
            will give better results, but using a subsampled subset can be
            a fast way to find the (approximate) location of a good
            hyperparameter set.

    Returns:
        score (float): The NMLL associated with the best lambda-beta values found.
        best_lb (np.ndarray): The best lambda and beta values found on optimization.
    """
    bounds = init_bounds.copy()
    if np.exp(bounds[0,0]) < 1e-3:
        bounds[0,0] = np.log(1e-3)

    hparams = np.zeros((sigma_vals.shape[0] + 1))
    if hparams.shape[0] > 1:
        hparams[1:] = sigma_vals
    kernel.set_hyperparams(hparams, logspace = True)

    eigvals, eigvecs, y_trans_y, ndatapoints = get_eigvals(kernel, dataset,
                        subsample = subsample)

    for _ in range(n_cycles):
        lambda_, spacing = get_grid_pts(bounds, n_pts_per_dim, kernel.device)

        scoregrid = generate_scoregrid(kernel, eigvals, eigvecs,
                            lambda_, y_trans_y, ndatapoints)

        min_pt = scoregrid.argmin()
        best_score, best_lb = scoregrid[min_pt], np.log(float(lambda_[min_pt]))
        bounds[0,0] = max(best_lb - spacing, init_bounds[0,0])
        bounds[0,1] = min(best_lb + spacing, init_bounds[0,1])

    return np.round(float(best_score), 3), np.round(np.asarray([best_lb]), 7)




def get_eigvals(kernel, dataset, subsample = 1):
    """Gets the eigenvalues and eigenvectors of the z_trans_z
    matrix. Note that this matrix may be ill-conditioned.
    These are required for the NMLL procedure, not for
    multifitting."""
    if kernel.device == "cuda":
        z_trans_z = cp.zeros((kernel.get_num_rffs(), kernel.get_num_rffs()))
        z_trans_y = cp.zeros((kernel.get_num_rffs()))
    else:
        z_trans_z = np.zeros((kernel.get_num_rffs(), kernel.get_num_rffs()))
        z_trans_y = np.zeros((kernel.get_num_rffs()))

    y_trans_y = 0.0
    ndatapoints = 0

    if subsample == 1:
        for xin, yin, ldata in dataset.get_chunked_data():
            xtrans, ydata = kernel.transform_x_y(xin, yin, ldata)
            z_trans_z += xtrans.T @ xtrans
            z_trans_y += xtrans.T @ ydata
            y_trans_y += ydata.T @ ydata
            ndatapoints += xin.shape[0]
    else:
        rng = np.random.default_rng(123)
        for xin, yin, ldata in dataset.get_chunked_data():
            idx_size = max(1, int(subsample * xin.shape[0]))
            idx = rng.choice(xin.shape[0], idx_size, replace=False)
            xin, yin, ldata = xin[idx,...], yin[idx], ldata[idx]
            xdata, ydata = kernel.transform_x(xin, yin, ldata)

            z_trans_y += xdata.T @ ydata
            z_trans_z += xdata.T @ xdata
            y_trans_y += ydata.T @ ydata
            ndatapoints += xdata.shape[0]

    z_trans_z.flat[::z_trans_z.shape[0]+1] += 1e-5
    if kernel.device == "cuda":
        eigvecs, eigvals, _ = cp.linalg.svd(z_trans_z, full_matrices = False)
    else:
        eigvecs, eigvals, _ = np.linalg.svd(z_trans_z, full_matrices = False)

    #Subtract out the additional 1e-5 for numerical stability.
    eigvals -= 1e-5

    mask = eigvals >= 1e-7
    cut_point = max(mask.sum(), 1)
    eigvals[cut_point:] = 1e-7
    eigvecs[:,cut_point:] = 0
    eigvecs = eigvecs.T @ z_trans_y
    return eigvals, eigvecs, y_trans_y, ndatapoints




def generate_scoregrid(kernel, eigvals, eigvecs, lambda_,
                    y_trans_y, ndatapoints):
    """Generates scores for all the lambda / beta values that
    are part of the current grid, using NMLL exact scoring.

    Args:
        dataset: A valid dataset object that can generate chunked data.
        kernel: A valid kernel object that can generate random features.
        eigvals (array): A (x) array of eigenvalues.
        eigvecs (array): A (x) array of eigenvectors.T @ (z^T @ y).
        lambda_ (array): A (num_params) array of lambda values.
        beta_ (array): A (num_params) array of beta values.
        y_trans_y (float): The quantity y^T y.

    Returns:
        scoregrid (array): A (num_params) array of scores.
    """
    eigval_batch = eigvals[:,None] + lambda_[None,:]**2
    scoregrid = y_trans_y - eigvecs.T @ (eigvecs[:,None] / eigval_batch)

    #The best possible for y^T Z (lambda**2 + Z^T Z)^-1 Z^T y is when
    #Z (lambda**2 + Z^T Z)^-1 Z^T y = Zw = y, i.e. the fit is perfect,
    #in which case y^T y - y^T Z (lambda**2 + Z^T Z)^-1 Z^T y = 0.
    #It is possible to get negative values due to rounding error for
    #highly ill-conditioned matrices, hence, we clip at zero.
    scoregrid[scoregrid < 0] = 0

    scoregrid *= 0.5
    if kernel.device == "cpu":
        logfun = np.log
        beta = np.sqrt(2 * scoregrid / (ndatapoints * lambda_**2))
    else:
        logfun = cp.log
        beta = cp.sqrt(2 * scoregrid / (ndatapoints * lambda_**2))

    #Allowing beta to take ANY value does not affect numerical stability but misrepresents
    #the outcome that would occur if a particular lambda_ value were selected.
    beta = beta.clip(min=0.1, max=10)

    scoregrid  /= (beta * lambda_)**2
    scoregrid += 0.5 * logfun(eigval_batch).sum(axis=0)

    scoregrid += (ndatapoints - kernel.get_num_rffs()) * logfun(lambda_)
    scoregrid += ndatapoints * 0.5 * np.log(2 * np.pi) + ndatapoints * logfun(beta)
    if kernel.device == "cuda":
        scoregrid = cp.asnumpy(scoregrid)
    return scoregrid


def get_grid_pts(bounds, n_pts_per_dim, device = "cpu"):
    """Generates grid points for the shared hyperparameters
    for a given set of boundaries. Used for both multifitting and
    NMLL scoring.

    Args:
        bounds (np.ndarray): A 1 x 2 array where [0,:] is the boundaries
            for shared hyperparameter lambda.
        n_pts_per_dim (int): The number of points per grid dimension.
        device (str): Either "cpu" or "cuda". Indicates where the values should
            be generated.

    Returns:
        lambda_ (np.ndarray): The grid coordinates for the first hyperparameter.
        spacing (array): The spacing between grid points.
    """
    lambda_pts = np.exp(np.linspace(bounds[0,0], bounds[0,1], n_pts_per_dim))

    spacing = 1.05 * np.abs(bounds[0,0] - bounds[0,1]) / n_pts_per_dim
    if device == "cuda":
        lambda_pts = cp.asarray(lambda_pts)
    return lambda_pts, spacing



"""Implements functions for fitting (once hyperparameters have been
selected) using CG, either using our internal routine or Scipy / Cupy's."""
import warnings
import numpy as np
from ..cg_toolkit.cg_tools import CPU_ConjugateGrad

try:
    import cupy as cp
    from ..cg_toolkit.cg_tools import GPU_ConjugateGrad
except:
    pass

from ..scoring_toolkit.exact_nmll_calcs import calc_zty




def cg_fit_lib_internal(kernel, dataset, cg_tol = 1e-4, max_iter = 500,
                        preconditioner = None, verbose = True):
    """Calculates the weights when fitting the model using
    preconditioned CG. Good scaling but slower for small
    numbers of random features.

    Args:
        kernel: A valid kernel object that can generate random features.
        dataset: Either OnlineDataset or OfflineDataset.
        cg_tol (float): The threshold below which cg is deemed to have
            converged. Defaults to 1e-5.
        max_iter (int): The maximum number of iterations before
            CG is deemed to have failed to converge.
        preconditioner: Either None or a valid Preconditioner (e.g.
            CudaRandomizedPreconditioner, CPURandomizedPreconditioner
            etc). If None, no preconditioning is used. Otherwise,
            the preconditioner is used for CG. The preconditioner
            can be built by calling self.build_preconditioner
            with appropriate arguments.
    verbose (bool): If True, print regular updates.

    Returns:
        weights: A cupy or numpy array of shape (M) for M
            random features.
        n_iter (int): The number of CG iterations.
        losses (list): The loss on each iteration; for diagnostic
            purposes.
    """
    if kernel.device == "cuda":
        cg_operator = GPU_ConjugateGrad()
        resid = cp.zeros((kernel.get_num_rffs(), 2, 1))
    else:
        cg_operator = CPU_ConjugateGrad()
        resid = np.zeros((kernel.get_num_rffs(), 2, 1))

    z_trans_y, _ = calc_zty(dataset, kernel)
    resid[:,0,:] = z_trans_y[:,None] / dataset.get_ndatapoints()

    weights, converged, n_iter, losses = cg_operator.fit(dataset, kernel,
                preconditioner, resid, max_iter, cg_tol, verbose,
                nmll_settings = False)
    weights *= dataset.get_ndatapoints()
    if not converged:
        warnings.warn("Conjugate gradients failed to converge! Try refitting "
                        "the model with updated settings.")

    if verbose:
        print(f"CG iterations: {n_iter}")
    return weights, n_iter, losses



"""Tests the crude tuning algorithm to ensure
we get performance on par with expectations."""
import sys
import os
import unittest

sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".."))
from utils.build_test_dataset import build_test_dataset
from utils.model_constructor import get_models


class CheckCrudeTuning(unittest.TestCase):
    """Tests the crude tuning algorithm."""

    def test_crude_tuning(self):
        """Test the crude tuning algorithm using an
        RBF kernel for simplicity (tuning & fitting with other
        kernels is tested under the complete pipeline tests."""
        online_data, _ = build_test_dataset(conv_kernel = False)
        cpu_mod, gpu_mod = get_models("RBF", online_data)
        cpu_mod.num_rffs = 512
        _, _, best_score = cpu_mod.tune_hyperparams_crude(online_data)
        self.assertTrue(best_score < 430)

        if gpu_mod is not None:
            gpu_mod.num_rffs = 512
            _, _, best_score = gpu_mod.tune_hyperparams_crude(online_data)
            self.assertTrue(best_score < 430)


if __name__ == "__main__":
    unittest.main()
